#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import pickle
import dill
from sqlalchemy import Column, Index, Integer, BigInteger
from sqlalchemy.orm import Session

from airflow.models.base import Base
from airflow.utils.session import provide_session


class EventProgress(Base):
    """the class to save the progress of the event."""

    __tablename__ = "event_progress"

    scheduling_job_id = Column(Integer, primary_key=True)
    last_event_time = Column(BigInteger, nullable=True)
    last_event_version = Column(BigInteger, nullable=True)

    def __init__(self, scheduling_job_id, last_event_time=None, last_event_version=None):
        self.scheduling_job_id = scheduling_job_id
        self.last_event_time = last_event_time
        self.last_event_version = last_event_version

    @provide_session
    def update_progress(self, scheduling_job_id: int,
                        last_event_time: int,
                        last_event_version: int,
                        session: Session = None):
        self.scheduling_job_id = scheduling_job_id
        self.last_event_time = last_event_time
        self.last_event_version = last_event_version
        session.merge(self)
        session.commit()


@provide_session
def create_or_update_progress(scheduling_job_id: int,
                              last_event_time: int,
                              last_event_version: int,
                              session: Session = None):
    progress = EventProgress(scheduling_job_id, last_event_time, last_event_version)
    session.merge(progress)
    session.commit()


@provide_session
def get_event_progress(scheduling_job_id: int,
                       session: Session = None):
    return session.query(EventProgress).filter(EventProgress.scheduling_job_id == scheduling_job_id).first()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import shutil
import unittest
import os
from unittest import mock
from ai_flow.plugin_interface.blob_manager_interface import BlobManagerFactory, BlobConfig

_TMP_FOLDER = '/tmp/' + __name__
_TMP_FILE = os.path.join(_TMP_FOLDER, 'file.txt')


class TestHDFSBlobManager(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        if not os.path.exists(_TMP_FOLDER):
            os.mkdir(_TMP_FOLDER)

    @classmethod
    def tearDownClass(cls) -> None:
        if os.path.exists(_TMP_FOLDER):
            shutil.rmtree(_TMP_FOLDER)

    def setUp(self) -> None:
        file = open(_TMP_FILE, 'w')
        file.close()

    def tearDown(self) -> None:
        if os.path.exists(_TMP_FILE):
            os.remove(_TMP_FILE)

    def test_without_root_directory_set(self):
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.hdfs_blob_manager.HDFSBlobManager'
        }
        blob_config = BlobConfig(config)
        with self.assertRaisesRegex(Exception, '`root_directory` option of blob manager config is not configured'):
            BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                   blob_config.blob_manager_config())

    @unittest.skipUnless((os.environ.get('blob_server.hdfs_url') is not None
                          and os.environ.get('blob_server.hdfs_user') is not None
                          and os.environ.get('blob_server.repo_name') is not None), 'need set hdfs config')
    def test_project_upload_download_hdfs(self):
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.hdfs_blob_manager.HDFSBlobManager',
            'blob_manager_config': {
                'hdfs_url': os.environ.get('blob_server.hdfs_url'),
                'hdfs_user': os.environ.get('blob_server.hdfs_user'),
                'root_directory': os.environ.get('blob_server.repo_name')
            }
        }
        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())
        uploaded_path = blob_manager.upload(_TMP_FILE)
        self.assertEqual(os.path.join(os.environ.get('blob_server.repo_name'), os.path.basename(_TMP_FILE)),
                         uploaded_path)

        downloaded_path = blob_manager.download(uploaded_path, _TMP_FOLDER)
        self.assertEqual(os.path.join(_TMP_FOLDER, os.path.basename(_TMP_FILE)), downloaded_path)

    def test_download_existed_file(self):
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.hdfs_blob_manager.HDFSBlobManager',
            'blob_manager_config': {
                'hdfs_url': 'mock',
                'hdfs_user': 'mock',
                'root_directory': 'hdfs:///path'
            }
        }
        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())

        def mock_download_file_from_hdfs(hdfs_path, local_path):
            with open(local_path, 'w') as f:
                f.write(str(time.time() * 1000))

        blob_manager._download_file_from_hdfs = mock_download_file_from_hdfs
        with mock.patch(
                'hdfs.client.InsecureClient'):
            blob_manager.download('hdfs:///path/file1', _TMP_FOLDER)
            blob_manager.download('hdfs:///path/file2', _TMP_FOLDER)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import time
import unittest
from threading import Thread
from unittest import mock

from ai_flow.workflow.job import Job

from ai_flow.workflow.status import Status

from ai_flow import DatasetMeta
from ai_flow.ai_graph.ai_node import AINode, ReadDatasetNode, WriteDatasetNode

from ai_flow.workflow.job_config import JobConfig

from ai_flow.ai_graph.ai_graph import AISubGraph
from ai_flow_plugins.job_plugins.bash import BashProcessor
from ai_flow_plugins.job_plugins.read_only import ReadOnlyProcessor, ReadOnlyJobGenerator, ReadOnlyJob, \
    ReadOnlyJobController, ReadOnlyJobHandle


class TestReadOnlyJobGenerator(unittest.TestCase):

    def test_generate_throw_unknown_type_exception(self):
        sub_graph = AISubGraph(JobConfig())
        ai_node = AINode(processor=BashProcessor('hello'))
        sub_graph.add_node(ai_node)
        sub_graph.add_node(AINode(processor=ReadOnlyProcessor()))
        job_generator = ReadOnlyJobGenerator()

        with self.assertRaises(TypeError):
            job_generator.generate(sub_graph)

    def test_generate(self):
        sub_graph = AISubGraph(JobConfig())
        sub_graph.add_node(ReadDatasetNode(dataset=DatasetMeta("test"), processor=ReadOnlyProcessor()))
        sub_graph.add_node(AINode(processor=ReadOnlyProcessor()))
        sub_graph.add_node(WriteDatasetNode(dataset=DatasetMeta("test"), processor=ReadOnlyProcessor()))
        job_generator = ReadOnlyJobGenerator()

        job = job_generator.generate(sub_graph)
        self.assertIsInstance(job, ReadOnlyJob)

    def test_generate_with_required_configs(self):
        job_config = JobConfig()
        sub_graph = AISubGraph(job_config)
        sub_graph.add_node(ReadDatasetNode(dataset=DatasetMeta("test"), processor=ReadOnlyProcessor()))
        sub_graph.add_node(AINode(processor=ReadOnlyProcessor()))
        sub_graph.add_node(WriteDatasetNode(dataset=DatasetMeta("test"), processor=ReadOnlyProcessor()))
        job_generator = ReadOnlyJobGenerator(required_properties={'required_key'})

        with self.assertRaises(RuntimeError):
            job_generator.generate(sub_graph)

        job_config.properties['required_key'] = 'value'
        job = job_generator.generate(sub_graph)
        self.assertIsInstance(job, ReadOnlyJob)


class TestReadOnlyJobController(unittest.TestCase):

    def setUp(self) -> None:
        self.job_controller = ReadOnlyJobController()
        self.job = ReadOnlyJob(JobConfig("test_job"))

    def test_submit_job(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info
        handle = self.job_controller.submit_job(self.job, job_runtime_env)
        self.assertIsInstance(handle, ReadOnlyJobHandle)
        self.assertEqual(self.job, handle.job)
        self.assertEqual(job_execution_info, handle.job_execution)

    def test_stop_job(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info
        with self.assertRaises(RuntimeError):
            self.job_controller.stop_job(ReadOnlyJobHandle(mock.Mock(), job_execution_info), job_runtime_env)

        handle = self.job_controller.submit_job(self.job, job_runtime_env)
        self.assertFalse(self.job_controller._job_stop_events[handle.job].is_set())
        self.job_controller.stop_job(handle, job_runtime_env)
        self.assertTrue(self.job_controller._job_stop_events[handle.job].is_set())

    def test_get_result(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info
        with self.assertRaises(RuntimeError):
            self.job_controller.get_result(ReadOnlyJobHandle(mock.Mock(), job_execution_info), job_runtime_env)

        handle = self.job_controller.submit_job(self.job, job_runtime_env)
        self.assertIsNone(self.job_controller.get_result(handle, False))

        def get_result():
            result = self.job_controller.get_result(handle, True)
            self.assertIsNone(result)
            self.assertTrue(self.job_controller._job_stop_events[handle.job].is_set())

        t = Thread(target=get_result)
        t.start()

        time.sleep(0.5)
        self.job_controller.stop_job(handle, job_runtime_env)
        t.join()

    def test_get_job_status(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info
        with self.assertRaises(RuntimeError):
            self.job_controller.get_job_status(ReadOnlyJobHandle(mock.Mock(), job_execution_info))
        handle = self.job_controller.submit_job(self.job, job_runtime_env)
        self.assertEqual(Status.RUNNING, self.job_controller.get_job_status(handle))

    def test_obtain_job_label(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info
        with self.assertRaises(RuntimeError):
            self.job_controller.obtain_job_label(ReadOnlyJobHandle(mock.Mock(), job_execution_info))

        handle = self.job_controller.submit_job(self.job, job_runtime_env)
        self.assertEqual("", self.job_controller.obtain_job_label(handle))

    def test_obtain_job_label_check_job_type(self):
        job_runtime_env = mock.Mock()
        job_execution_info = mock.Mock()
        job_runtime_env.job_execution_info = job_execution_info

        job = Job(mock.Mock())
        handle = self.job_controller.submit_job(job, job_runtime_env)
        with self.assertRaises(TypeError):
            self.job_controller.obtain_job_label(handle)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from typing import Text, Set

from ai_flow.util.json_utils import Jsonable

from notification_service.base_notification import BaseEvent

WORKFLOW_EXECUTION_DEFAULT_CONTEXT = 'default'


class EventContext(ABC):
    """
    The context of the event returns by ContextExtractor.
    """

    @abstractmethod
    def is_broadcast(self) -> bool:
        """
        :return: Whether the event should be broadcast.
        """
        pass

    @abstractmethod
    def get_contexts(self) -> Set[Text]:
        """
        :return: A set of contexts the event belongs to.
        """
        pass


class Broadcast(EventContext):
    """
    This class indicates that the event should be broadcast.
    """

    def is_broadcast(self) -> bool:
        return True

    def get_contexts(self) -> Set[Text]:
        return set()


class ContextList(EventContext):
    """
    A set of context the event belongs to.

    Note: 'default' is a reserved key word for default context.
    """

    def __init__(self):
        self._contexts: Set[Text] = set()

    def is_broadcast(self) -> bool:
        return False

    def get_contexts(self) -> Set[Text]:
        return self._contexts

    def add_context(self, context: Text):
        self._contexts.add(context)


class ContextExtractor(ABC, Jsonable):
    """
    ContextExtractor can be implemented by user to decide if an event should be broadcast or we should extract context
    from an event. If the event should be broadcast, it will be handle by all the workflow executions and job executions
    of that workflow. Otherwise, only workflow execution and job executions with the same context can handle the event.
    """

    @abstractmethod
    def extract_context(self, event: BaseEvent) -> EventContext:
        """
        This method is called to decide if an event should be broadcast or we should extract context from the event.
        If the event should be broadcast, return :class:`Broadcast`. The event will be handle by all the workflow
        executions and job executions of that workflow. Otherwise, return :class:`ContextList`. Only workflow execution
        and job executions with the same context can handle the event.

        If a :class:`ContextList` with an empty set or None is returns, the event is dropped.

        :param event: The :class:`~notification_service.base_notification.BaseEvent` to extract context from.
        :return: The context of the event or it should be broadcast.
        """
        pass


class BroadcastAllContextExtractor(ContextExtractor):
    """
    BroadcastAllContextExtractor is the default ContextExtractor to used. It marks all events as broadcast events.
    """

    def extract_context(self, event: BaseEvent) -> EventContext:
        return Broadcast()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import unittest
import os
import shutil
from ai_flow.workflow.periodic_config import PeriodicConfig

from ai_flow import AIFlowServerRunner, init_ai_flow_context
import ai_flow as af
from ai_flow.workflow.control_edge import JobAction
from ai_flow.workflow.status import Status
from ai_flow.test.api.mock_plugins import MockJobFactory
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server


class TestDagGenerator(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ns_server = start_notification_server()
        config_file = os.path.dirname(os.path.dirname(os.path.dirname(__file__))) + '/master.yaml'
        cls.master = AIFlowServerRunner(config_file=config_file)
        cls.master.start()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.master.stop()
        stop_notification_server(cls.ns_server)

    def setUp(self):
        self.master._clear_db()
        af.current_graph().clear_graph()
        init_ai_flow_context()

    def tearDown(self):
        self.master._clear_db()
        generated = '{}/generated'.format(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
        if os.path.exists(generated):
            shutil.rmtree(generated)

    def test_one_task(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)

    def test_two_task(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        with af.job_config('task_2'):
            af.user_define_operation(processor=None)
        af.action_on_event(job_name='task_2', event_key='a', event_type='a', event_value='a', sender='task_1')
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue("op_1.subscribe_event('a', 'a', 'default', 'task_1')" in code)
        self.assertTrue("op_1.set_events_handler(AIFlowHandler(configs_op_1))" in code)

    def test_three_task(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        with af.job_config('task_2'):
            af.user_define_operation(processor=None)
        with af.job_config('task_3'):
            af.user_define_operation(processor=None)
        af.action_on_event(job_name='task_3', event_key='a', event_type='a', event_value='a', sender='task_1')
        af.action_on_job_status(job_name='task_3', upstream_job_name='task_2',
                                upstream_job_status=Status.FINISHED,
                                action=JobAction.START)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue(".subscribe_event('a', 'a', 'default', 'task_1')" in code)
        # Now do not support the event_type equals JOB_STATUS_CHANGED event.
        # self.assertTrue(".subscribe_event('test_dag_generator', 'JOB_STATUS_CHANGED', 'test_project', 'task_2')" in code)
        self.assertTrue(".set_events_handler(AIFlowHandler(configs_op_" in code)

    def test_periodic_cron_task(self):
        with af.job_config('task_4'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('op_0.executor_config' in code)

    def test_periodic_interval_task(self):
        with af.job_config('task_5'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')

        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('op_0.executor_config' in code)

    def test_periodic_cron_workflow(self):
        workflow_config_ = af.current_workflow_config()
        workflow_config_.periodic_config = PeriodicConfig(trigger_config={'start_date': "2020,1,1,,,,Asia/Chongqing",
                                                                          'cron': "*/5 * * * * * *"})
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('datetime' in code.split('default_args')[1])
        self.assertTrue('schedule_interval' in code.split('default_args')[1])
        self.assertTrue('utc' not in code.split('default_args')[1])

    def test_periodic_cron_workflow_without_start_date(self):
        workflow_config_ = af.current_workflow_config()
        workflow_config_.periodic_config = PeriodicConfig(trigger_config={'cron': "*/5 * * * * * *"})
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('datetime.fromisoformat' in code.split('default_args')[1])
        self.assertTrue('schedule_interval' in code.split('default_args')[1])
        self.assertTrue('utc' not in code.split('default_args')[1])

    def test_periodic_cron_workflow_with_timezone(self):
        workflow_config_ = af.current_workflow_config()
        workflow_config_.periodic_config = PeriodicConfig(trigger_config={'start_date': "2020,1,1,,,,",
                                                                          'cron': "*/5 * * * * * *", 'timezone': 'utc'})
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('datetime' in code.split('default_args')[1])
        self.assertTrue('schedule_interval' in code.split('default_args')[1])
        self.assertTrue('utc' in code.split('default_args')[1])

    def test_periodic_interval_workflow(self):
        workflow_config_ = af.current_workflow_config()
        workflow_config_.periodic_config = PeriodicConfig(trigger_config={'start_date': "2020,1,1,,,,Asia/Chongqing",
                                                                          'interval': "1,1,1,"})
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('datetime' in code)
        self.assertTrue('schedule_interval' in code)
        self.assertTrue('timedelta' in code)

    def test_periodic_interval_workflow_without_start_date(self):
        workflow_config_ = af.current_workflow_config()
        workflow_config_.periodic_config = PeriodicConfig(trigger_config={'interval': "1,1,1,"})
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('datetime.fromisoformat' in code)
        self.assertTrue('schedule_interval' in code)
        self.assertTrue('timedelta' in code)

    def test_action_on_job_status(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        with af.job_config('task_2'):
            af.user_define_operation(processor=None)
        with af.job_config('task_3'):
            af.user_define_operation(processor=None)
        af.action_on_job_status(job_name='task_2', upstream_job_name='task_1')
        af.action_on_job_status(job_name='task_3', upstream_job_name='task_2',
                                upstream_job_status=Status.RUNNING,
                                action=JobAction.START)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue(
            "\nop_1.subscribe_event('test_dag_generator.task_1', 'TASK_STATUS_CHANGED', 'test_project', 'task_1')" in code)
        self.assertTrue(
            "\nop_2.subscribe_event('test_dag_generator.task_2', 'TASK_STATUS_CHANGED', 'test_project', 'task_2')" in code)

    def test_action_on_job_status_two_status(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        with af.job_config('task_2'):
            af.user_define_operation(processor=None)
        af.action_on_job_status(job_name='task_2', upstream_job_name='task_1',
                                upstream_job_status=Status.RUNNING,
                                action=JobAction.START)
        af.action_on_job_status(job_name='task_2', upstream_job_name='task_1', upstream_job_status=Status.FINISHED,
                                action=JobAction.STOP)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('"event_value": "RUNNING"' in code)
        self.assertTrue('"event_value": "FINISHED"' in code)

    def test_generate_dag_with_context_extractor(self):
        with af.job_config('task_1'):
            af.user_define_operation(processor=None)
        with af.job_config('task_2'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        print(code)
        self.assertTrue('\nimport os' in code)
        self.assertTrue('\nfrom ai_flow_plugins.scheduler_plugins.airflow.context_extractor '
                        'import AIFlowContextExtractorAdaptor' in code)
        self.assertTrue('\ncontext_extractor_base64_str = ' in code)
        self.assertTrue('\ndag.context_extractor = AIFlowContextExtractorAdaptor(context_extractor_base64_str)' in code)

    def test_one_task_with_airflow_args(self):
        with af.job_config('task_6'):
            af.user_define_operation(processor=None)
        w = af.workflow_operation.submit_workflow(workflow_name='test_dag_generator')
        code = w.properties.get('code')
        self.assertTrue('op_0 = AIFlowOperator' in code)
        self.assertTrue('retries=2' in code)
        self.assertTrue('retry_delay=timedelta(seconds=5)' in code)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import datetime
import logging
import os
import shutil
from abc import ABC
from tempfile import NamedTemporaryFile
from typing import Dict, Text, List, Optional

import cloudpickle
from airflow.exceptions import DagNotFound

from ai_flow_plugins.scheduler_plugins.airflow.dag_generator import DAGGenerator
from ai_flow.context.project_context import ProjectContext
from ai_flow.plugin_interface.scheduler_interface import Scheduler, \
    WorkflowInfo, JobExecutionInfo, WorkflowExecutionInfo
from ai_flow_plugins.scheduler_plugins.airflow.airflow_restful_util import AirFlowRestfulUtil
from ai_flow.workflow.workflow import Workflow
from ai_flow.workflow import status
from ai_flow.util.time_utils import datetime_to_int64
from airflow.executors.scheduling_action import SchedulingAction
from airflow.models.taskexecution import TaskExecution
from airflow.models.dag import DagModel
from airflow.models.dagrun import DagRun
from airflow.utils.db import create_session
from airflow.utils.state import State
from airflow.contrib.jobs.scheduler_client import EventSchedulerClient, SCHEDULER_NAMESPACE, ExecutionContext
from airflow.utils.dates import parse_execution_date
from airflow.api.common.experimental import delete_dag

logger = logging.getLogger(__name__)

TIMEOUT = 5 * 60


class AirFlowSchedulerBase(Scheduler, ABC):
    """
    AirFlowScheduler is an implementation of a Scheduler interface based on AirFlow.
    AirFlowScheduler contains two configuration items:
    1. notification_server_uri: The address of NotificationService.
    2. airflow_deploy_path: AirFlow dag file deployment directory.
    """

    def __init__(self, config: Dict):
        if 'notification_server_uri' not in config:
            raise Exception('`notification_server_uri` option of scheduler config is not configured. '
                            'Please add the `notification_server_uri` option under `scheduler_config` option!')
        if 'airflow_deploy_path' not in config:
            from airflow import settings
            dags_folder = settings.DAGS_FOLDER
            config['airflow_deploy_path'] = dags_folder
            logger.info("airflow_deploy_path is set to {} from airflow config".format(dags_folder))
        super().__init__(config)
        self.dag_generator = DAGGenerator()
        self._airflow_client = None

    @classmethod
    def airflow_dag_id(cls, namespace, workflow_name):
        return '{}.{}'.format(namespace, workflow_name)

    @classmethod
    def dag_id_to_namespace_workflow(cls, dag_id: Text):
        tmp = dag_id.split('.')
        return tmp[0], tmp[1]

    @classmethod
    def generate_workflow_execution_id(cls, dag_id, run_id) -> Text:
        return '{}|{}'.format(dag_id, run_id)

    @classmethod
    def generate_job_execution_id(cls, dag_id, run_id, task_id, seq_num) -> Text:
        return '{}|{}|{}|{}'.format(dag_id, run_id, task_id, seq_num)

    @classmethod
    def parse_dag_id_and_run_id(cls, workflow_execution_id: Text) -> (Text, Text):
        return workflow_execution_id.split('|')

    @classmethod
    def airflow_state_to_status(cls, state) -> status.Status:
        if State.SUCCESS == state:
            return status.Status.FINISHED
        elif State.FAILED == state:
            return status.Status.FAILED
        elif State.RUNNING == state:
            return status.Status.RUNNING
        elif State.KILLED == state or State.SHUTDOWN == state \
                or State.KILLING == state:
            # We map airflow state KILLING to KILLED in the assumption that KILLING is a transient state,
            # and it is the best we can do.
            return status.Status.KILLED
        else:
            return status.Status.INIT

    @classmethod
    def status_to_airflow_state(cls, status_: status.Status) -> Text:
        if status.Status.FINISHED == status_:
            return State.SUCCESS
        elif status.Status.FAILED == status_:
            return State.FAILED
        elif status.Status.RUNNING == status_:
            return State.RUNNING
        elif status.Status.KILLED == status_:
            return State.KILLED
        else:
            return State.NONE

    @property
    def airflow_client(self):
        if self._airflow_client is None:
            self._airflow_client = EventSchedulerClient(
                notification_server_uri=self.config.get('notification_server_uri'),
                namespace=SCHEDULER_NAMESPACE)
        return self._airflow_client

    def submit_workflow(self, workflow: Workflow, context_extractor, project_context: ProjectContext) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_context.project_name, workflow.workflow_name)
        self._check_configurable_path('resource_dir')
        code_text = self.dag_generator.generate(workflow=workflow,
                                                project_name=project_context.project_name,
                                                resource_dir=self.config.get('resource_dir'),
                                                base_log_folder=self.config.get('base_log_folder'),
                                                context_extractor=context_extractor)
        self._check_configurable_path('airflow_deploy_path')
        airflow_file_path = self._write_to_deploy_path(code_text, dag_id + ".py",
                                                       self.config.get('airflow_deploy_path'))

        self.airflow_client.trigger_parse_dag(airflow_file_path, timeout=TIMEOUT)
        return WorkflowInfo(namespace=project_context.project_name,
                            workflow_name=workflow.workflow_name,
                            properties={'dag_file': airflow_file_path})

    @staticmethod
    def _write_to_deploy_path(content, filename, deploy_path, mode='w+t'):
        airflow_file_path = os.path.join(deploy_path, filename)
        if os.path.exists(airflow_file_path):
            os.remove(airflow_file_path)
        with NamedTemporaryFile(mode=mode, prefix=filename, dir='/tmp', delete=False) as f:
            f.write(content)
        shutil.move(f.name, airflow_file_path)
        return airflow_file_path

    def delete_workflow(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        deploy_path = self.config.get('airflow_deploy_path')
        if deploy_path is None:
            raise Exception("airflow_deploy_path config not set!")

        self.pause_workflow_scheduling(project_name=project_name,
                                       workflow_name=workflow_name)
        self.stop_all_workflow_execution(project_name=project_name,
                                         workflow_name=workflow_name)
        try:
            delete_dag.delete_dag(dag_id)
        except DagNotFound:
            raise Exception("DAG with id {} not found. Cannot delete".format(dag_id))

        airflow_file_path = os.path.join(deploy_path,
                                         dag_id + '.py')
        if os.path.exists(airflow_file_path):
            os.remove(airflow_file_path)
            return WorkflowInfo(namespace=project_name,
                                workflow_name=workflow_name,
                                properties={'dag_file': airflow_file_path})
        else:
            return None

    def _check_configurable_path(self, config_str):
        config_path = self.config.get(config_str)
        if config_path is None:
            raise Exception("{} config not set!".format(config_str))
        if not os.path.exists(config_path):
            os.makedirs(config_path)

    def stop_scheduling_job(self, workflow_execution_id: Text, job_name: Text):
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        self.airflow_client.stop_scheduling_task(dag_id=dag_id,
                                                 dagrun_id=run_id,
                                                 task_id=job_name,
                                                 timeout=TIMEOUT)

    def resume_scheduling_job(self, workflow_execution_id: Text, job_name: Text):
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        self.airflow_client.resume_scheduling_task(dag_id=dag_id,
                                                   dagrun_id=run_id,
                                                   task_id=job_name,
                                                   timeout=TIMEOUT)


class AirFlowScheduler(AirFlowSchedulerBase):
    """
    AirFlowScheduler is an implementation of a Scheduler interface based on AirFlow.
    AirFlowScheduler contains two configuration items:
    1. notification_server_uri: The address of server of NotificationService.
    2. airflow_deploy_path: AirFlow dag file deployment directory.
    """

    def __init__(self, config: Dict):
        super().__init__(config)

    @classmethod
    def dag_exist(cls, dag_id):
        with create_session() as session:
            dag = session.query(DagModel).filter(DagModel.dag_id == dag_id).first()
            if dag is None:
                return False
            else:
                return True

    def pause_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        DagModel.get_dagmodel(dag_id=dag_id).set_is_paused(is_paused=True)
        return WorkflowInfo(namespace=project_name, workflow_name=workflow_name)

    def resume_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        DagModel.get_dagmodel(dag_id=dag_id).set_is_paused(is_paused=False)
        return WorkflowInfo(namespace=project_name, workflow_name=workflow_name)

    def start_new_workflow_execution(self, project_name: Text, workflow_name: Text,
                                     workflow_execution_context: Text = None) \
            -> Optional[WorkflowExecutionInfo]:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        deploy_path = self.config.get('airflow_deploy_path')
        if deploy_path is None:
            raise Exception("airflow_deploy_path config not set!")
        if not self.dag_exist(dag_id):
            return None
        context: ExecutionContext = self.airflow_client.schedule_dag(dag_id, workflow_execution_context, timeout=TIMEOUT)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(session=session, dag_id=dag_id, run_id=context.dagrun_id)
            if dagrun is None:
                return None
            else:
                return WorkflowExecutionInfo(
                    workflow_info=WorkflowInfo(namespace=project_name, workflow_name=workflow_name),
                    workflow_execution_id=self.generate_workflow_execution_id(dag_id=dag_id, run_id=dagrun.run_id),
                    status=status.Status.INIT)

    def stop_all_workflow_execution(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        workflow_execution_list = self.list_workflow_executions(project_name, workflow_name)
        for we in workflow_execution_list:
            if we.status == status.Status.RUNNING:
                self.stop_workflow_execution(we.workflow_execution_id)
        return workflow_execution_list

    def stop_workflow_execution_by_context(self, workflow_name: Text, context: Text) -> Optional[WorkflowExecutionInfo]:
        # TODO: impl
        pass

    def stop_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
            if dagrun.state in State.finished:
                return WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                        workflow_name=workflow_name),
                                             workflow_execution_id=workflow_execution_id,
                                             status=self.airflow_state_to_status(dagrun.state))
            context: ExecutionContext = ExecutionContext(dagrun_id=dagrun.run_id)
            current_context = self.airflow_client.stop_dag_run(dagrun.dag_id, context, timeout=TIMEOUT)
            return WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                    workflow_name=workflow_name),
                                         workflow_execution_id=workflow_execution_id,
                                         status=status.Status.KILLED)

    def get_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            else:
                status_ = self.airflow_state_to_status(dagrun.state)
                project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
                return WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                        workflow_name=workflow_name),
                                             workflow_execution_id=workflow_execution_id,
                                             status=status_,
                                             start_date=str(datetime_to_int64(dagrun.start_date)),
                                             end_date=str(datetime_to_int64(dagrun.end_date))
                                             )

    def list_workflow_executions(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        with create_session() as session:
            dagrun_list = session.query(DagRun).filter(DagRun.dag_id == dag_id).all()
            if dagrun_list is None:
                return []
            else:
                result = []
                for dagrun in dagrun_list:
                    status_ = self.airflow_state_to_status(dagrun.state)
                    result.append(WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                   workflow_name=workflow_name),
                                                        workflow_execution_id
                                                        =self.generate_workflow_execution_id(dag_id, dagrun.run_id),
                                                        status=status_,
                                                        start_date=str(datetime_to_int64(dagrun.start_date)),
                                                        end_date=str(datetime_to_int64(dagrun.end_date)),
                                                        ))
                return result

    def start_job_execution(self, job_name: Text, workflow_execution_id: Text) -> Optional[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            if dagrun.state != State.RUNNING:
                raise Exception('execution: {} state: {} can not trigger job.'.format(workflow_execution_id,
                                                                                      dagrun.state))
            task = dagrun.get_task_instance(job_name, session)
            if task is None:
                return None
            if task.state in State.unfinished:
                raise Exception('job:{} state: {} can not start!'.format(job_name, task.state))
            self.airflow_client.schedule_task(dag_id=dagrun.dag_id,
                                              task_id=job_name,
                                              action=SchedulingAction.START,
                                              context=ExecutionContext(dagrun_id=dagrun.run_id),
                                              timeout=TIMEOUT)
            project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
            return JobExecutionInfo(job_name=job_name,
                                    status=self.airflow_state_to_status(task.state),
                                    workflow_execution
                                    =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                      workflow_name=workflow_name),
                                                           workflow_execution_id=workflow_execution_id,
                                                           status=self.airflow_state_to_status(dagrun.state)))

    def stop_job_execution(self, job_name: Text, workflow_execution_id: Text) -> Optional[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            task = dagrun.get_task_instance(job_name, session)
            if task is None:
                return None
            if task.state in State.finished:
                raise Exception('job:{} state: {} can not stop!'.format(job_name, task.state))
            else:
                self.airflow_client.schedule_task(dag_id=dagrun.dag_id,
                                                  task_id=job_name,
                                                  action=SchedulingAction.STOP,
                                                  context=ExecutionContext(dagrun_id=dagrun.run_id),
                                                  timeout=TIMEOUT)
            project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
            return JobExecutionInfo(job_name=job_name,
                                    status=self.airflow_state_to_status(task.state),
                                    workflow_execution
                                    =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                      workflow_name=workflow_name),
                                                           workflow_execution_id=workflow_execution_id,
                                                           status=self.airflow_state_to_status(dagrun.state)))

    def restart_job_execution(self, job_name: Text, workflow_execution_id: Text) -> Optional[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            if dagrun.state != State.RUNNING:
                raise Exception('execution: {} state: {} can not trigger job.'.format(workflow_execution_id,
                                                                                      dagrun.state))
            task = dagrun.get_task_instance(job_name, session)
            if task is None:
                return None
            self.airflow_client.schedule_task(dag_id=dagrun.dag_id,
                                              task_id=job_name,
                                              action=SchedulingAction.RESTART,
                                              context=ExecutionContext(dagrun_id=dagrun.run_id),
                                              timeout=TIMEOUT)
            project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
            return JobExecutionInfo(job_name=job_name,
                                    status=self.airflow_state_to_status(task.state),
                                    workflow_execution
                                    =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                      workflow_name=workflow_name),
                                                           workflow_execution_id=workflow_execution_id,
                                                           status=self.airflow_state_to_status(dagrun.state)))

    def get_job_executions(self, job_name: Text, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            task_list = session.query(TaskExecution).filter(TaskExecution.dag_id == dagrun.dag_id,
                                                            TaskExecution.execution_date == dagrun.execution_date,
                                                            TaskExecution.task_id == job_name).all()
            if task_list is None:
                return []
            else:
                result = self.build_job_execution_info_list(dagrun, task_list)
                return result

    def build_job_execution_info_list(self, dagrun, task_list):
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.dag_id)
        result = []
        for task in task_list:
            job = JobExecutionInfo(job_execution_id=self.generate_job_execution_id(dag_id=dagrun.dag_id,
                                                                                   run_id=dagrun.run_id,
                                                                                   task_id=task.task_id,
                                                                                   seq_num=task.seq_num),
                                   job_name=task.task_id,
                                   status=self.airflow_state_to_status(task.state),
                                   start_date=str(datetime_to_int64(task.start_date)),
                                   end_date=str(datetime_to_int64(task.end_date)),
                                   execution_label=task.execution_label,
                                   workflow_execution
                                   =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                     workflow_name=workflow_name),
                                                          workflow_execution_id=self.generate_workflow_execution_id(
                                                              dag_id=dagrun.dag_id,
                                                              run_id=dagrun.run_id),
                                                          status=self.airflow_state_to_status(dagrun.state)))
            result.append(job)
        return result

    def list_job_executions(self, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        with create_session() as session:
            dagrun = DagRun.get_run_by_id(dag_id=dag_id, run_id=run_id, session=session)
            if dagrun is None:
                return None
            task_list = session.query(TaskExecution).filter(TaskExecution.dag_id == dagrun.dag_id,
                                                            TaskExecution.execution_date == dagrun.execution_date).all()
            if task_list is None:
                return []
            else:
                result = self.build_job_execution_info_list(dagrun, task_list)
                return result


class AirFlowSchedulerRestful(AirFlowSchedulerBase):

    def __init__(self, config: Dict):
        super().__init__(config)
        if 'endpoint_url' not in config:
            raise Exception('`endpoint_url` option of scheduler config is not configured. '
                            'Please add the `endpoint_url` option under `scheduler_config` option!')
        if 'username' not in config:
            raise Exception('`username` option of scheduler config is not configured. '
                            'Please add the `username` option under `scheduler_config` option!')
        if 'password' not in config:
            raise Exception('`password` option of scheduler config is not configured. '
                            'Please add the `password` option under `scheduler_config` option!')
        self.restful_util: AirFlowRestfulUtil = AirFlowRestfulUtil(endpoint_url=config.get("endpoint_url"),
                                                                   user_name=config.get("username"),
                                                                   password=config.get("password"))

    @classmethod
    def datetime_str_to_int64_str(cls, datetime_str):
        if datetime_str is None:
            return '0'
        else:
            return str(datetime_to_int64(parse_execution_date(datetime_str)))

    def pause_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        self.restful_util.set_dag_is_paused(dag_id=dag_id, is_paused=True)
        return WorkflowInfo(namespace=project_name, workflow_name=workflow_name)

    def resume_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        self.restful_util.set_dag_is_paused(dag_id=dag_id, is_paused=False)
        return WorkflowInfo(namespace=project_name, workflow_name=workflow_name)

    def start_new_workflow_execution(self, project_name: Text,
                                     workflow_name: Text,
                                     context: Text = None) -> Optional[WorkflowExecutionInfo]:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        deploy_path = self.config.get('airflow_deploy_path')
        if deploy_path is None:
            raise Exception("airflow_deploy_path config not set!")
        if not self.restful_util.dag_exist(dag_id):
            return None
        context: ExecutionContext = self.airflow_client.schedule_dag(dag_id, context, timeout=TIMEOUT)
        dagrun = self.restful_util.get_dagrun(dag_id=dag_id, run_id=context.dagrun_id)
        if dagrun is None:
            return None
        else:
            return WorkflowExecutionInfo(
                workflow_info=WorkflowInfo(namespace=project_name, workflow_name=workflow_name),
                workflow_execution_id=self.generate_workflow_execution_id(dag_id, context.dagrun_id),
                status=status.Status.INIT)

    def stop_all_workflow_execution(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        dagrun_list = self.restful_util.list_dagruns(dag_id)
        result = []
        for dagrun in dagrun_list:
            status_ = self.airflow_state_to_status(dagrun.get('state'))
            workflow_execution_id = self.generate_workflow_execution_id(dag_id, dagrun.get('dag_run_id'))
            if status_ == State.RUNNING:
                self.stop_workflow_execution(workflow_execution_id)
            result.append(WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                           workflow_name=workflow_name),
                                                workflow_execution_id=workflow_execution_id,
                                                status=status_,
                                                start_date=self.datetime_str_to_int64_str(dagrun.get('start_date')),
                                                end_date=self.datetime_str_to_int64_str(dagrun.get('end_date'))
                                                ))
        return result

    def stop_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id=dag_id, run_id=run_id)
        if dagrun is None:
            return None
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dag_id)
        context: ExecutionContext = ExecutionContext(dagrun_id=run_id)
        current_context = self.airflow_client.stop_dag_run(dag_id, context, timeout=TIMEOUT)
        return WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                workflow_name=workflow_name),
                                     workflow_execution_id=workflow_execution_id,
                                     status=status.Status.KILLED)

    def stop_workflow_execution_by_context(self, workflow_name: Text, context: Text) -> Optional[WorkflowExecutionInfo]:
        # todo Need to implement the function
        raise NotImplementedError('Does not implement the top_workflow_execution_by_context function.')

    def get_workflow_execution(self, workflow_execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id=dag_id, run_id=run_id)
        print(dagrun)
        if dagrun is None:
            return None
        else:
            status_ = self.airflow_state_to_status(dagrun.get('state'))
            project_name, workflow_name = self.dag_id_to_namespace_workflow(dag_id)
            return WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                    workflow_name=workflow_name),
                                         workflow_execution_id=workflow_execution_id,
                                         status=status_,
                                         start_date=self.datetime_str_to_int64_str(dagrun.get('start_date')),
                                         end_date=self.datetime_str_to_int64_str(dagrun.get('end_date'))
                                         )

    def list_workflow_executions(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        dag_id = self.airflow_dag_id(project_name, workflow_name)
        dagrun_list = self.restful_util.list_dagruns(dag_id)
        result = []
        for dagrun in dagrun_list:
            status_ = self.airflow_state_to_status(dagrun.get('state'))
            workflow_execution_id = self.generate_workflow_execution_id(dag_id, dagrun.get('dag_run_id'))
            result.append(WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                           workflow_name=workflow_name),
                                                workflow_execution_id=workflow_execution_id,
                                                status=status_,
                                                start_date=self.datetime_str_to_int64_str(dagrun.get('start_date')),
                                                end_date=self.datetime_str_to_int64_str(dagrun.get('end_date')),
                                                ))
        return result

    def start_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id, run_id)
        if dagrun is None:
            return None
        if dagrun.get('state') != State.RUNNING:
            raise Exception('execution: {} state: {} can not trigger job.'.format(workflow_execution_id,
                                                                                  dagrun.get('state')))
        task = self.restful_util.get_task_instance(dag_id, run_id, job_name)
        if task is None:
            return None
        if task.get('state') in State.running:
            raise Exception('job:{} state: {} can not start!'.format(job_name, task.get('state')))
        self.airflow_client.schedule_task(dag_id=dag_id,
                                          task_id=job_name,
                                          action=SchedulingAction.START,
                                          context=ExecutionContext(dagrun_id=run_id),
                                          timeout=TIMEOUT)
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dag_id)
        return JobExecutionInfo(job_name=job_name,
                                status=self.airflow_state_to_status(task.get('state')),
                                workflow_execution
                                =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                  workflow_name=workflow_name),
                                                       workflow_execution_id=workflow_execution_id,
                                                       status=self.airflow_state_to_status(dagrun.get('state'))))

    def stop_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id, run_id)
        if dagrun is None:
            return None
        task = self.restful_util.get_task_instance(dag_id, run_id, job_name)
        if task is None:
            return None
        if task.get('state') in State.finished:
            raise Exception('job:{} state: {} can not stop!'.format(job_name, task.get('state')))
        else:
            self.airflow_client.schedule_task(dag_id=dag_id,
                                              task_id=job_name,
                                              action=SchedulingAction.STOP,
                                              context=ExecutionContext(dagrun_id=run_id),
                                              timeout=TIMEOUT)
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dag_id)
        return JobExecutionInfo(job_name=job_name,
                                status=self.airflow_state_to_status(task.get('state')),
                                workflow_execution
                                =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                  workflow_name=workflow_name),
                                                       workflow_execution_id=workflow_execution_id,
                                                       status=self.airflow_state_to_status(dagrun.get('state'))))

    def restart_job_execution(self, job_name: Text, workflow_execution_id: Text) -> JobExecutionInfo:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id, run_id)
        if dagrun is None:
            return None
        if dagrun.get('state') != State.RUNNING:
            raise Exception('execution: {} state: {} can not trigger job.'.format(workflow_execution_id,
                                                                                  dagrun.get('state')))
        task = self.restful_util.get_task_instance(dag_id, run_id, job_name)
        if task is None:
            return None
        self.airflow_client.schedule_task(dag_id=dag_id,
                                          task_id=job_name,
                                          action=SchedulingAction.RESTART,
                                          context=ExecutionContext(dagrun_id=run_id),
                                          timeout=TIMEOUT)
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dag_id)
        return JobExecutionInfo(job_name=job_name,
                                status=self.airflow_state_to_status(task.get('state')),
                                workflow_execution
                                =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                  workflow_name=workflow_name),
                                                       workflow_execution_id=workflow_execution_id,
                                                       status=self.airflow_state_to_status(dagrun.get('state'))))

    def build_job_execution_info_list(self, dagrun, task_list):
        project_name, workflow_name = self.dag_id_to_namespace_workflow(dagrun.get('dag_id'))
        workflow_execution_id = self.generate_workflow_execution_id(dagrun.get('dag_id'), dagrun.get('dag_run_id'))
        result = []
        for task in task_list:
            job = JobExecutionInfo(job_execution_id=self.generate_job_execution_id(dag_id=dagrun.get('dag_id'),
                                                                                   run_id=dagrun.get('dag_run_id'),
                                                                                   task_id=task.get('task_id'),
                                                                                   seq_num=task.get('seq_num')),
                                   job_name=task.get('task_id'),
                                   status=self.airflow_state_to_status(task.get('state')),
                                   start_date=self.datetime_str_to_int64_str(task.get('start_date')),
                                   end_date=self.datetime_str_to_int64_str(task.get('end_date')),
                                   execution_label=task.execution_label,
                                   workflow_execution
                                   =WorkflowExecutionInfo(workflow_info=WorkflowInfo(namespace=project_name,
                                                                                     workflow_name=workflow_name),
                                                          workflow_execution_id=workflow_execution_id,
                                                          status=self.airflow_state_to_status(dagrun.get('state'))))
            result.append(job)
        return result

    def get_job_executions(self, job_name: Text, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id, run_id)
        if dagrun is None:
            return None
        tasks = self.restful_util.list_task_execution_by_task_id(dag_id, run_id, job_name)
        if tasks is None:
            return []
        else:
            return self.build_job_execution_info_list(dagrun, tasks)

    def list_job_executions(self, workflow_execution_id: Text) -> List[JobExecutionInfo]:
        dag_id, run_id = self.parse_dag_id_and_run_id(workflow_execution_id)
        dagrun = self.restful_util.get_dagrun(dag_id, run_id)
        if dagrun is None:
            return None
        task_list = self.restful_util.list_task_execution(dag_id, run_id)
        if task_list is None:
            return []
        else:
            result = self.build_job_execution_info_list(dagrun, task_list)
            return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_aiflow_member

Revision ID: de045f309236
Revises: de1c96ef582a
Create Date: 2021-10-13 16:48:26.412700

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'de045f309236'
down_revision = 'de1c96ef582a'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('aiflow_member',
                    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
                    sa.Column('version', sa.BigInteger(), nullable=False),
                    sa.Column('server_uri', sa.String(length=767), nullable=False),
                    sa.Column('update_time', sa.BigInteger(), nullable=False),
                    sa.Column('uuid', sa.String(length=128), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('server_uri'),
                    sa.UniqueConstraint('uuid')
                    )


def downgrade():
    op.drop_table('aiflow_member')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import sys
from typing import List, Callable

from ai_flow.meta.workflow_meta import WorkflowMeta
from ai_flow.plugin_interface.scheduler_interface import Scheduler
from ai_flow.scheduler_service.service.workflow_event_handler import WorkflowEventHandler
from ai_flow.scheduler_service.service.workflow_execution_event_handler_state import WorkflowContextEventHandlerState
from ai_flow.store.abstract_store import AbstractStore
from ai_flow.workflow.control_edge import WorkflowAction
from ai_flow.workflow.status import Status
from notification_service.base_notification import BaseEvent, ANY_CONDITION

WorkflowEventHandlerFactoryType = Callable[[WorkflowMeta], WorkflowEventHandler]


class EventKey:
    def __init__(self, namespace, event_type, key, sender):
        self.namespace = namespace
        self.event_type = event_type
        self.key = key
        self.sender = sender


class Poison(object):
    pass


class WorkflowEventProcessor:
    def __init__(self, conn, store: AbstractStore, scheduler: Scheduler,
                 workflow_event_handler_factory: WorkflowEventHandlerFactoryType = WorkflowEventHandler):
        super().__init__()
        self._conn = conn
        self.scheduler = scheduler
        self._workflow_event_handler_factory = workflow_event_handler_factory
        self.store = store

    def run(self) -> None:
        while True:
            try:
                event = self._conn.recv()
                if isinstance(event, Poison):
                    break
                elif isinstance(event, BaseEvent):
                    self._process_event(event)
                else:
                    logging.error("Receive unknown type of event: {}".format(type(event)))
            except EOFError as _:
                logging.info("Event channel is closed exiting")
                break
            except Exception as e:
                logging.error("Unexpected Exception", exc_info=e)

    def _process_event(self, event: BaseEvent):
        from ai_flow import ProjectMeta
        projects: List[ProjectMeta] = self.store.list_projects(sys.maxsize, 0)
        if projects is None:
            return
        for project in projects:
            project_name = project.name
            workflows: List[WorkflowMeta] = self.store.list_workflows(project_name=project_name)
            if workflows is None:
                continue
            for workflow in workflows:
                if workflow.last_event_version is not None and workflow.last_event_version >= event.version:
                    logging.info("Workflow {} last processed event version is {}, skipping event: {}"
                                 .format(workflow.name, workflow.last_event_version, event))
                    continue

                if self._is_subscribed(workflow, event):
                    self._handle_event_for_workflow(project_name, workflow, event)

                self._update_workflow_last_event_version(project_name, workflow, event)

    def _update_workflow_last_event_version(self, project_name, workflow, event):
        self.store.update_workflow(workflow.name, project_name,
                                   context_extractor_in_bytes=workflow.context_extractor_in_bytes,
                                   scheduling_rules=workflow.scheduling_rules,
                                   last_event_version=event.version)

    def _is_subscribed(self, workflow: WorkflowMeta, event: BaseEvent):
        rules = workflow.scheduling_rules
        subscribed_event_keys = [EventKey(e.namespace, e.event_type, e.event_key, e.sender)
                                 for rule in rules for e in rule.event_condition.events]
        event_key = EventKey(event.namespace, event.event_type, event.key, event.sender)
        return self._is_event_key_subscribed(event_key, subscribed_event_keys)

    @staticmethod
    def _is_event_key_subscribed(event_key: EventKey, subscribed_event_keys: List[EventKey]) -> bool:
        def match_condition(config_value, event_value) -> bool:
            if config_value == ANY_CONDITION or event_value == config_value:
                return True
            else:
                return False

        for subscribed_event_key in subscribed_event_keys:
            c_namespace = match_condition(subscribed_event_key.namespace, event_key.namespace)
            c_event_type = match_condition(subscribed_event_key.event_type, event_key.event_type)
            c_sender = match_condition(subscribed_event_key.sender, event_key.sender)
            c_key = match_condition(subscribed_event_key.key, event_key.key)
            if c_namespace and c_event_type and c_sender and c_key:
                return True
        return False

    def _handle_event_for_workflow(self, project_name, workflow: WorkflowMeta, event: BaseEvent):
        try:
            event_context = workflow.get_context_extractor().extract_context(event)
        except Exception as e:
            logging.error("Failed to call context extractor, workflow {} skips event {}".format(workflow.name, event),
                          exc_info=e)
            return
        states = self._get_workflow_execution_state(event_context, project_name, workflow.name)

        for state in states:
            event_handler = self._workflow_event_handler_factory(workflow)
            action, new_state = event_handler.handle_event(event, workflow_state=state.state)
            if action == WorkflowAction.START:
                workflow_execution = None
                if state.workflow_execution_id is not None:
                    workflow_execution = self.scheduler.get_workflow_execution(state.workflow_execution_id)

                if workflow_execution is None or workflow_execution.status != Status.RUNNING:
                    workflow_execution = self.scheduler.start_new_workflow_execution(state.project_name,
                                                                                     state.workflow_name,
                                                                                     state.context)
                if workflow_execution is not None:
                    state.workflow_execution_id = workflow_execution.workflow_execution_id
                else:
                    logging.warning("Fail to start workflow execution for project: {} workflow: {} context: {}"
                                    .format(state.project_name, state.workflow_name, state.context))
            elif action == WorkflowAction.STOP:
                # only stop if workflow execution exist
                if state.workflow_execution_id is not None:
                    self.scheduler.stop_workflow_execution(state.workflow_execution_id)

            state.state = new_state
            self.store.update_workflow_context_event_handler_state(state.project_name, state.workflow_name,
                                                                   state.context, state.workflow_execution_id,
                                                                   state.state)

    def _get_workflow_execution_state(self, event_context,
                                      project_name, workflow_name) -> List[WorkflowContextEventHandlerState]:
        states: List[WorkflowContextEventHandlerState] = []
        if event_context.is_broadcast():
            states.extend(self.store.list_workflow_context_event_handler_states(project_name, workflow_name))
        else:
            contexts = event_context.get_contexts()
            for context in contexts:
                state = self.store.get_workflow_context_event_handler_state(project_name, workflow_name, context)

                # register state for context if it is not exist
                if state is None:
                    state = self.store.register_workflow_context_event_handler_state(project_name,
                                                                                     workflow_name, context)
                states.append(state)
        return states



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class BashProcessor(object):
    """
    BashProcessor is the processor corresponding to the bash job.
    """
    def __init__(self,
                 bash_command: str,
                 output_encoding: str = 'utf-8'):
        """
        :param bash_command: The command of the bash job.
        :param output_encoding: The bash job output encoding type.
        """
        self.bash_command = bash_command
        self.output_encoding = output_encoding



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import os
import traceback
from typing import Text

from ai_flow.api.context_extractor import WORKFLOW_EXECUTION_DEFAULT_CONTEXT
from ai_flow.context.project_context import ProjectContext, build_project_context
from ai_flow.endpoint.server.workflow_proto_utils import workflow_to_proto, workflow_execution_to_proto, \
    workflow_execution_list_to_proto, job_to_proto, job_list_to_proto
from ai_flow.plugin_interface.blob_manager_interface import BlobConfig, BlobManagerFactory
from ai_flow.plugin_interface.scheduler_interface import Scheduler, SchedulerFactory, WorkflowInfo
from ai_flow.protobuf.message_pb2 import ResultProto, StatusProto
from ai_flow.protobuf.scheduling_service_pb2 import \
    (ScheduleWorkflowRequest,
     WorkflowInfoResponse,
     WorkflowExecutionRequest,
     WorkflowExecutionResponse,
     ListWorkflowExecutionResponse,
     ScheduleJobRequest,
     JobInfoResponse,
     ListJobInfoResponse, WorkflowExecutionOnEventRequest)
from ai_flow.protobuf.scheduling_service_pb2_grpc import SchedulingServiceServicer
from ai_flow.scheduler_service.service.config import SchedulerServiceConfig
from ai_flow.scheduler_service.service.workflow_event_manager import WorkflowEventManager
from ai_flow.store.db.db_util import create_db_store
from ai_flow.util import json_utils
from ai_flow.util.file_util.zip_file_util import extract_zip_file
from ai_flow.workflow.control_edge import WorkflowAction
from ai_flow.workflow.workflow import Workflow
from ai_flow.workflow.workflow import WorkflowPropertyKeys


class SchedulerService(SchedulingServiceServicer):
    def __init__(self,
                 scheduler_service_config: SchedulerServiceConfig,
                 db_uri,
                 notification_server_uri,
                 base_log_folder):
        self._scheduler_service_config = scheduler_service_config
        scheduler_service_config.scheduler().scheduler_config().update({'base_log_folder': base_log_folder})
        self._scheduler: Scheduler \
            = SchedulerFactory.create_scheduler(scheduler_service_config.scheduler().scheduler_class(),
                                                scheduler_service_config.scheduler().scheduler_config())
        self.store = create_db_store(db_uri)

        if notification_server_uri is not None:
            self.workflow_event_manager = WorkflowEventManager(notification_server_uri=notification_server_uri,
                                                               db_uri=db_uri,
                                                               scheduler_service_config=scheduler_service_config)
        else:
            self.workflow_event_manager = None
            logging.warning('notification_server_uri is None, workflow event manager did not start. '
                            'WorkflowExecution will not start/stop on events')

    def start(self):
        if self.workflow_event_manager:
            self.workflow_event_manager.start()

    def stop(self):
        logging.info("stopping SchedulerService.")
        if self.workflow_event_manager:
            self.workflow_event_manager.stop()
        logging.info("SchedulerService stopped.")

    # workflow interface
    def submitWorkflow(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            if rq.workflow_json is None or '' == rq.workflow_json:
                return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                               error_message='workflow json is empty!'))
            workflow: Workflow = json_utils.loads(rq.workflow_json)
            raw_config = {}
            raw_config.update(workflow.properties[WorkflowPropertyKeys.BLOB])
            blob_config = BlobConfig(raw_config)
            blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                                  blob_config.blob_manager_config())
            project_zip_path = blob_manager.download(remote_file_path=workflow.project_uri,
                                                     local_dir=self._scheduler_service_config.repository())
            extract_path = os.path.splitext(project_zip_path)[0]
            project_path: Text = extract_zip_file(zip_file_path=project_zip_path,
                                                  extract_path=extract_path)
            project_context: ProjectContext = build_project_context(project_path)
            project_name = project_context.project_name

            workflow_meta = \
                self.store.get_workflow_by_name(project_name=project_name, workflow_name=rq.workflow_name)
            workflow_info = self._scheduler.submit_workflow(workflow, workflow_meta.get_context_extractor(),
                                                            project_context)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{}, {} do not exist!'.format(project_name,
                                                                                   workflow.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def deleteWorkflow(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.delete_workflow(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def pauseWorkflowScheduling(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.pause_workflow_scheduling(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def resumeWorkflowScheduling(self, request, context):
        try:
            rq: ScheduleWorkflowRequest = request
            workflow_info = self._scheduler.resume_workflow_scheduling(rq.namespace, rq.workflow_name)
            if workflow_info is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.workflow_name)))
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def getWorkflow(self, request, context):
        return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                       error_message='Do not support getWorkflow'))

    def listWorkflows(self, request, context):
        return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                       error_message='Do not support listWorkflows'))

    # workflow execution interface

    def startNewWorkflowExecutionOnEvent(self, request, context):
        try:
            rq: WorkflowExecutionOnEventRequest = request
            namespace = rq.namespace
            workflow_name = rq.workflow_name
            event_conditions_json = rq.event_conditions_json
            workflow = self.store.get_workflow_by_name(namespace, workflow_name)
            workflow.update_condition(json_utils.loads(event_conditions_json), WorkflowAction.START)
            self.store.update_workflow(project_name=namespace, workflow_name=workflow_name,
                                       context_extractor_in_bytes=workflow.context_extractor_in_bytes,
                                       scheduling_rules=workflow.scheduling_rules)
            context_state = self.store \
                .get_workflow_context_event_handler_state(project_name=namespace,
                                                          workflow_name=workflow_name,
                                                          context=WORKFLOW_EXECUTION_DEFAULT_CONTEXT)
            if context_state is None:
                self.store.register_workflow_context_event_handler_state(
                    project_name=namespace,
                    workflow_name=workflow_name,
                    context=WORKFLOW_EXECUTION_DEFAULT_CONTEXT)

            workflow_info = WorkflowInfo(namespace=namespace, workflow_name=workflow_name)
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                           error_message=traceback.format_exc()))

    def startNewWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            context = rq.context.value if rq.HasField('context') else None
            workflow_execution = self._scheduler.start_new_workflow_execution(rq.namespace, rq.workflow_name,
                                                                              context)
            if workflow_execution is None:
                return WorkflowInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{}, {} do not exist!'.format(rq.namespace, rq.workflow_name)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def killAllWorkflowExecutions(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution_list = self._scheduler.stop_all_workflow_execution(rq.namespace, rq.workflow_name)
            response = ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK))
            response.workflow_execution_list.extend(workflow_execution_list_to_proto(workflow_execution_list))
            return response
        except Exception as err:
            return ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                    error_message=traceback.format_exc()))

    def killWorkflowExecutionOnEvent(self, request, context):
        try:
            rq: WorkflowExecutionOnEventRequest = request
            namespace = rq.namespace
            workflow_name = rq.workflow_name
            event_conditions_json = rq.event_conditions_json
            workflow = self.store.get_workflow_by_name(namespace, workflow_name)
            workflow.update_condition(json_utils.loads(event_conditions_json), WorkflowAction.STOP)
            self.store.update_workflow(project_name=namespace, workflow_name=workflow_name,
                                       context_extractor_in_bytes=workflow.context_extractor_in_bytes,
                                       scheduling_rules=workflow.scheduling_rules)
            context_state = self.store \
                .get_workflow_context_event_handler_state(project_name=namespace,
                                                          workflow_name=workflow_name,
                                                          context=WORKFLOW_EXECUTION_DEFAULT_CONTEXT)
            if context_state is None:
                self.store.register_workflow_context_event_handler_state(
                    project_name=namespace,
                    workflow_name=workflow_name,
                    context=WORKFLOW_EXECUTION_DEFAULT_CONTEXT)

            workflow_info = WorkflowInfo(namespace=namespace, workflow_name=workflow_name)
            return WorkflowInfoResponse(result=ResultProto(status=StatusProto.OK),
                                        workflow=workflow_to_proto(workflow_info))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def killWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution = self._scheduler.stop_workflow_execution(rq.execution_id)
            if workflow_execution is None:
                return WorkflowExecutionResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.execution_id)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def getWorkflowExecution(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution = self._scheduler.get_workflow_execution(rq.execution_id)
            if workflow_execution is None:
                return WorkflowExecutionResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.execution_id)))
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK),
                                             workflow_execution=workflow_execution_to_proto(workflow_execution))
        except Exception as err:
            return WorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                error_message=traceback.format_exc()))

    def listWorkflowExecutions(self, request, context):
        try:
            rq: WorkflowExecutionRequest = request
            workflow_execution_list = self._scheduler.list_workflow_executions(rq.namespace, rq.workflow_name)
            response = ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.OK))
            response.workflow_execution_list.extend(workflow_execution_list_to_proto(workflow_execution_list))
            return response
        except Exception as err:
            return ListWorkflowExecutionResponse(result=ResultProto(status=StatusProto.ERROR,
                                                                    error_message=traceback.format_exc()))

    # job interface
    def startJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.start_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def stopJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.stop_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def restartJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job = self._scheduler.restart_job_execution(rq.job_name, rq.execution_id)
            if job is None:
                return JobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            return JobInfoResponse(result=ResultProto(status=StatusProto.OK), job=job_to_proto(job))
        except Exception as err:
            return JobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                      error_message=traceback.format_exc()))

    def getJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            jobs = self._scheduler.get_job_executions(rq.job_name, rq.execution_id)
            if jobs is None:
                return ListJobInfoResponse(
                    result=ResultProto(status=StatusProto.ERROR,
                                       error_message='{} do not exist!'.format(rq.job_name)))
            else:
                response = ListJobInfoResponse(result=ResultProto(status=StatusProto.OK))
                response.job_list.extend(job_list_to_proto(jobs))
                return response
        except Exception as err:
            return ListJobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                          error_message=traceback.format_exc()))

    def listJobs(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            job_list = self._scheduler.list_job_executions(rq.execution_id)
            response = ListJobInfoResponse(result=ResultProto(status=StatusProto.OK))
            response.job_list.extend(job_list_to_proto(job_list))
            return response
        except Exception as err:
            return ListJobInfoResponse(result=ResultProto(status=StatusProto.ERROR,
                                                          error_message=traceback.format_exc()))

    def stopSchedulingJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            self._scheduler.stop_scheduling_job(workflow_execution_id=rq.execution_id,
                                                job_name=rq.job_name)
            response = ResultProto(status=StatusProto.OK)
            return response
        except Exception as err:
            return ResultProto(status=StatusProto.ERROR, error_message=traceback.format_exc())

    def resumeSchedulingJob(self, request, context):
        try:
            rq: ScheduleJobRequest = request
            self._scheduler.resume_scheduling_job(workflow_execution_id=rq.execution_id,
                                                  job_name=rq.job_name)
            response = ResultProto(status=StatusProto.OK)
            return response
        except Exception as err:
            return ResultProto(status=StatusProto.ERROR, error_message=traceback.format_exc())




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Custom logging formatter for Airflow"""

import logging
from logging import StreamHandler

from airflow.configuration import conf
from airflow.utils.helpers import parse_template_string


class TaskHandlerWithCustomFormatter(StreamHandler):
    """Custom implementation of StreamHandler, a class which writes logging records for Airflow"""

    def __init__(self, stream):
        super().__init__()
        self.prefix_jinja_template = None

    def set_context(self, ti):
        """
        Accept the run-time context (i.e. the current task) and configure the formatter accordingly.

        :param ti:
        :return:
        """
        if ti.raw:
            return
        prefix = conf.get('logging', 'task_log_prefix_template')

        rendered_prefix = ""
        if prefix:
            _, self.prefix_jinja_template = parse_template_string(prefix)
            rendered_prefix = self._render_prefix(ti)
        formatter = logging.Formatter(rendered_prefix + ":" + self.formatter._fmt)  # pylint: disable=W0212
        self.setFormatter(formatter)
        self.setLevel(self.level)

    def _render_prefix(self, ti):
        if self.prefix_jinja_template:
            jinja_context = ti.get_template_context()
            return self.prefix_jinja_template.render(**jinja_context)
        logging.warning("'task_log_prefix_template' is in invalid format, ignoring the variable value")
        return ""



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging.config
import os

from ai_flow.endpoint.server.server_config import AIFlowServerConfig

# We hard code the logging config, we should make it configurable in the future.
logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '[%(asctime)s - %(filename)s:%(lineno)d [%(threadName)s] - %(levelname)s: %(message)s',
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
            'formatter': 'default'
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': ['console']
    }
})


def get_aiflow_home():
    if 'AIFLOW_HOME' in os.environ:
        home = os.getenv('AIFLOW_HOME')
    else:
        home = os.getenv('HOME') + '/aiflow'
    return home


AIFLOW_HOME = get_aiflow_home()
AIFLOW_PID_FILENAME = 'aiflow_server.pid'
AIFLOW_WEBSERVER_PID_FILENAME = "aiflow_web_server.pid"


def get_configuration_file_path():
    config_file_path = AIFLOW_HOME + '/aiflow_server.yaml'
    if not os.path.exists(config_file_path):
        raise FileNotFoundError('Config file {} not found.'.format(config_file_path))
    return config_file_path


def get_configuration():
    config_file_path = get_configuration_file_path()
    config = AIFlowServerConfig()
    config.load_from_file(config_file_path)
    return config



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import fcntl
import os
import time
from typing import Text, Dict, Any
from hdfs.client import InsecureClient
from ai_flow.plugin_interface.blob_manager_interface import BlobManager


class HDFSBlobManager(BlobManager):
    """
    HDFSBlobManager is an implementation of BlobManager based on the hdfs file system
    HDFSBlobManager contains configuration items:
    1. hdfs_url: Hostname or IP address of HDFS namenode, prefixed with protocol, followed by WebHDFS port on namenode
    2. hdfs_user: User default. Defaults to the current user's (as determined by `whoami`).
    3. root_directory: The upload directory of the project.
    """

    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        if not self.root_dir:
            raise Exception('`root_directory` option of blob manager config is not configured.')
        hdfs_url = config.get('hdfs_url', None)
        hdfs_user = config.get('hdfs_user', 'default')
        hdfs_client = InsecureClient(url=hdfs_url, user=hdfs_user)
        self._hdfs_client = hdfs_client

    def upload(self, local_file_path: Text) -> Text:
        """
        Upload a given file to blob server. Uploaded file will be placed under self.root_dir.

        :param local_file_path: the path of file to be uploaded.
        :return the uri of the uploaded file in blob server.
        """
        file_name = os.path.basename(local_file_path)
        remote_file_path = os.path.join(self.root_dir, file_name)
        self._hdfs_client.upload(hdfs_path=remote_file_path, local_path=local_file_path)
        return remote_file_path

    def download(self, remote_file_path: Text, local_dir: Text) -> Text:
        """
        Download file from remote blob server to local directory.
        Only files located in self.root_dir can be downloaded by BlobManager.

        :param remote_file_path: The path of file to be downloaded.
        :param local_dir: the local directory.
        :return the local uri of the downloaded file.
        """
        self._check_remote_path_legality(remote_file_path)
        file_name = os.path.basename(remote_file_path)
        local_file_path = os.path.join(local_dir, file_name)

        if not os.path.exists(local_file_path):
            lock_file_path = "{}.lock".format(local_file_path)
            lock_file = open(lock_file_path, 'w')
            fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX)
            try:
                if not os.path.exists(local_file_path):
                    self.log.info("Downloading file from HDFS: {}".format(remote_file_path))
                    self._download_file_from_hdfs(hdfs_path=remote_file_path, local_path=local_file_path)
            except Exception as e:
                self.log.error("Failed to download file: {}".format(remote_file_path), exc_info=e)
            finally:
                fcntl.flock(lock_file.fileno(), fcntl.LOCK_UN)
                self.log.debug('Unlocked file {}'.format(lock_file_path))
                lock_file.close()
                if os.path.exists(lock_file_path):
                    try:
                        os.remove(lock_file_path)
                    except OSError as e:
                        self.log.warning("Failed to remove lock file: {}".format(lock_file_path), exc_info=e)
        else:
            self.log.info("HDFS file: {} already exist at {}".format(remote_file_path, local_file_path))
        return local_file_path

    def _download_file_from_hdfs(self, hdfs_path, local_path, retry_sleep_sec=5):
        for i in range(3):
            try:
                self._hdfs_client.download(hdfs_path=hdfs_path, local_path=local_path)
                return
            except Exception as e:
                self.log.error("Downloading file {} failed, retrying {}/3 in {} second".format(
                    hdfs_path, i+1, retry_sleep_sec), exc_info=e)
                time.sleep(retry_sleep_sec)
        raise RuntimeError("Failed to download HDFS file: {}".format(hdfs_path))

    def _check_remote_path_legality(self, file_path: Text):
        """
        Check if the file can be downloaded by blob manager.

        :param file_path: The path of file to be checked.
        """
        if not file_path.startswith(self.root_dir):
            raise Exception("Cannot download {} from blob server".format(file_path))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import socket


def get_ip_addr():
    """
    Get ip address of localhost by UDP socket.
    :return: ip address of localhost
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    except OSError:
        ip = '127.0.0.1'
    finally:
        s.close()
        return ip



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import json
import unittest
from unittest import mock

from ai_flow.workflow.job_config import JobConfig

from ai_flow_plugins.job_plugins.flink.read_only_flink.read_only_flink_job_plugin import ReadOnlyFlinkJobController
from ai_flow_plugins.job_plugins.read_only import ReadOnlyJob


class TestReadOnlyFlinkJobController(unittest.TestCase):
    def test_get_job_label(self):
        controller = ReadOnlyFlinkJobController()
        mock_response = """
Waiting for response...
------------------ Running/Restarting Jobs -------------------
15.09.2021 10:12:54 : b219545e02782aeaf14cce56615af4b7 : CarTopSpeedWindowingExample (RUNNING)
--------------------------------------------------------------
No scheduled jobs.
---------------------- Terminated Jobs -----------------------
15.09.2021 10:12:34 : 6a3f3c0520f89bda4132365489a12e60 : CarTopSpeedWindowingExample (CANCELED)
--------------------------------------------------------------
        """

        with mock.patch.object(ReadOnlyFlinkJobController, "_list_flink_job_status", return_value=mock_response):
            job = ReadOnlyJob(JobConfig("test_job", "read_only_flink", properties={'job_id': '12345'}))
            self.assertEqual("", controller.get_job_label(job))

            job = ReadOnlyJob(JobConfig("test_job", "read_only_flink",
                                        properties={'job_id': 'b219545e02782aeaf14cce56615af4b7'}))
            label = json.loads(controller.get_job_label(job))
            self.assertEqual("15.09.2021 10:12:54", label['start_time'])
            self.assertEqual("CarTopSpeedWindowingExample", label['job_name'])
            self.assertEqual("RUNNING", label['status'])

            job = ReadOnlyJob(JobConfig("test_job", "read_only_flink",
                                        properties={'job_id': '6a3f3c0520f89bda4132365489a12e60'}))
            label = json.loads(controller.get_job_label(job))
            self.assertEqual("15.09.2021 10:12:34", label['start_time'])
            self.assertEqual("CarTopSpeedWindowingExample", label['job_name'])
            self.assertEqual("CANCELED", label['status'])




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import io
from contextlib import redirect_stdout

from ai_flow.cli.commands.job_command import job_list_executions, job_restart_execution, job_show_execution, \
    job_start_execution, job_stop_execution, job_stop_scheduling, job_resume_scheduling
from ai_flow.test.cli.commands.test_command import TestCommand, PROJECT_PATH


class TestCliJob(TestCommand):

    def test_cli_job_list_executions(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_list_executions(
                self.parser.parse_args(['job', 'list-executions', PROJECT_PATH, '1']))
        self.assertEquals(8, len(str.splitlines(stdout.getvalue())[0].split('|')))

    def test_cli_job_restart_execution(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_restart_execution(
                self.parser.parse_args(['job', 'restart-execution', PROJECT_PATH, 'task_1', '1']))
        self.assertEquals('Job: {}, workflow execution: {}, restarted: {}.'.format('task_1', '1', True),
                          str.splitlines(stdout.getvalue())[0])

    def test_cli_job_show_execution(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_show_execution(
                self.parser.parse_args(['job', 'show-execution', PROJECT_PATH, 'task_1', '1']))
        self.assertEquals(8, len(str.splitlines(stdout.getvalue())[0].split('|')))

    def test_cli_job_start_execution(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_start_execution(
                self.parser.parse_args(['job', 'start-execution', PROJECT_PATH, 'task_1', '1']))
        self.assertEquals('Job: {}, workflow execution: {}, started: {}.'.format('task_1', '1', True),
                          str.splitlines(stdout.getvalue())[0])

    def test_cli_job_stop_execution(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_stop_execution(
                self.parser.parse_args(['job', 'stop-execution', PROJECT_PATH, 'task_1', '1']))
        self.assertEquals('Job: {}, workflow execution: {}, stopped: {}.'.format('task_1', '1', True),
                          str.splitlines(stdout.getvalue())[0])

    def test_cli_job_stop_scheduling(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_stop_scheduling(
                self.parser.parse_args(['job', 'stop-scheduling', PROJECT_PATH, 'task_1', '1']))
        self.assertTrue("stop scheduling" in stdout.getvalue())

    def test_cli_job_resume_scheduling(self):
        with redirect_stdout(io.StringIO()) as stdout:
            job_resume_scheduling(
                self.parser.parse_args(['job', 'resume-scheduling', PROJECT_PATH, 'task_1', '1']))
        self.assertTrue("resume scheduling" in stdout.getvalue())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
import os
from ai_flow.ai_graph.ai_node import AINode, ReadDatasetNode, WriteDatasetNode
from ai_flow.api import ops
from ai_flow.ai_graph.ai_graph import current_graph
from ai_flow.api.ai_flow_context import init_ai_flow_context
from ai_flow.context.job_context import job_config
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.meta.dataset_meta import DatasetMeta
from ai_flow.meta.model_meta import ModelMeta
from ai_flow.test.util.server_util import wait_for_server_started
from ai_flow.workflow.control_edge import ControlEdge, JobAction, AIFlowInternalEventType
from ai_flow.workflow.periodic_config import PeriodicConfig
from ai_flow.workflow.status import Status
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server


_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'


class TestOps(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ns_server = start_notification_server()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        cls.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT,
                                  start_meta_service=True,
                                  start_metric_service=False,
                                  start_model_center_service=False,
                                  start_scheduler_service=False)
        cls.server.run()
        wait_for_server_started('localhost:{}'.format(_PORT))
        init_ai_flow_context()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.server.stop()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        stop_notification_server(cls.ns_server)

    def setUp(self):
        init_ai_flow_context()
        current_graph().clear_graph()

    def tearDown(self):
        current_graph().clear_graph()

    def get_node_by_name(self, name) -> AINode:
        for n in current_graph().nodes.values():
            if n.name == name:
                return n
        return None

    def test_user_define_operation(self):
        with job_config('task_1'):
            o = ops.user_define_operation(processor=None, a='a', name='1')
            ops.user_define_operation(input=o, b='b', name='2')
        self.assertEqual(2, len(current_graph().nodes))
        self.assertEqual(1, len(current_graph().edges))
        node_0 = list(current_graph().nodes.values())[0]
        node_1 = list(current_graph().nodes.values())[1]
        self.assertEqual('mock', node_0.config.job_type)
        self.assertEqual('mock', node_1.config.job_type)
        self.assertEqual('a', self.get_node_by_name('1').node_config.get('a'))
        self.assertEqual('b', self.get_node_by_name('2').node_config.get('b'))

    def test_read_write_dataset(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='source'))
            ops.write_dataset(input=o, dataset_info=DatasetMeta(name='sink'))
        self.assertEqual(2, len(current_graph().nodes))
        self.assertEqual(1, len(current_graph().edges))
        node_list = list(current_graph().nodes.values())
        for node in node_list:
            if isinstance(node, ReadDatasetNode):
                self.assertEqual('source', node.node_config.get('dataset').name)
            elif isinstance(node, WriteDatasetNode):
                self.assertEqual('sink', node.node_config.get('dataset').name)
            self.assertEqual('mock', node.config.job_type)

    def test_transform(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'))
            t = ops.transform(input=o, transform_processor=None)
            ops.write_dataset(input=t, dataset_info=DatasetMeta(name='dataset'))
        self.assertEqual(3, len(current_graph().nodes))
        self.assertEqual(2, len(current_graph().edges))

    def test_train(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'))
            t = ops.train(input=o, training_processor=None, output_num=1, model_info=ModelMeta(name='model'), name='a')
            ops.write_dataset(input=t, dataset_info=DatasetMeta(name='dataset'))
        self.assertEqual(3, len(current_graph().nodes))
        self.assertEqual(2, len(current_graph().edges))
        n = self.get_node_by_name('a')
        self.assertEqual('model', n.node_config.get('model_info').name)

    def test_predict(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'))
            t = ops.predict(input=o, prediction_processor=None,
                            model_info=ModelMeta(name='model'), name='a')
            ops.write_dataset(input=t, dataset_info=DatasetMeta(name='dataset'))
        self.assertEqual(3, len(current_graph().nodes))
        self.assertEqual(2, len(current_graph().edges))
        n = self.get_node_by_name('a')
        self.assertEqual('model', n.node_config.get('model_info').name)

    def test_evaluate(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'))
            t = ops.evaluate(input=o, evaluation_processor=None,
                             model_info=ModelMeta(name='model'), name='a')
        self.assertEqual(2, len(current_graph().nodes))
        self.assertEqual(1, len(current_graph().edges))
        n = self.get_node_by_name('a')
        self.assertEqual('model', n.node_config.get('model_info').name)

    def test_dataset_validate(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'), name='a')
            ops.dataset_validate(input=o, dataset_validation_processor=None, name='b')
        self.assertEqual(2, len(current_graph().nodes))
        self.assertEqual(1, len(current_graph().edges))
        n = self.get_node_by_name('a')
        self.assertEqual('dataset', n.node_config.get('dataset').name)

    def test_model_validate(self):
        with job_config('task_1'):
            o = ops.read_dataset(read_dataset_processor=None, dataset_info=DatasetMeta(name='dataset'))
            t = ops.model_validate(input=o, model_validation_processor=None,
                                   model_info=ModelMeta(name='model'), name='a')
        self.assertEqual(2, len(current_graph().nodes))
        self.assertEqual(1, len(current_graph().edges))
        n = self.get_node_by_name('a')
        self.assertEqual('model', n.node_config.get('model_info').name)

    def test_push_model(self):
        with job_config('task_1'):
            ops.push_model(pushing_model_processor=None, model_info=ModelMeta(name='model'), name='a')
        self.assertEqual(1, len(current_graph().nodes))
        n = self.get_node_by_name('a')
        self.assertEqual('model', n.node_config.get('model_info').name)

    def test_action_on_event(self):
        with job_config('task_1'):
            o1 = ops.user_define_operation(processor=None, a='a', name='1')
        with job_config('task_2'):
            o2 = ops.user_define_operation(processor=None, b='b', name='2')
        ops.action_on_event(job_name='task_1', sender='task_2', event_key='a', event_value='a')
        self.assertEqual(1, len(current_graph().edges))
        edge: ControlEdge = current_graph().edges.get('task_1')[0]
        self.assertEqual('task_1', edge.destination)
        events = edge.scheduling_rule.event_condition.events
        event = next(iter(events))
        self.assertEqual('task_2', event.sender)
        self.assertEqual('a', event.event_key)
        self.assertEqual('a', event.event_value)

    def test_action_on_status(self):
        with job_config('task_1'):
            o1 = ops.user_define_operation(processor=None, a='a', name='1')
        with job_config('task_2'):
            o2 = ops.user_define_operation(processor=None, b='b', name='2')
        ops.action_on_job_status(job_name='task_1',
                                 upstream_job_name='task_2',
                                 upstream_job_status=Status.FINISHED, action=JobAction.START)
        self.assertEqual(1, len(current_graph().edges))
        edge: ControlEdge = current_graph().edges.get('task_1')[0]
        self.assertEqual('task_1', edge.destination)

        events = edge.scheduling_rule.event_condition.events
        event = next(iter(events))
        self.assertEqual('task_2', event.sender)
        self.assertEqual(AIFlowInternalEventType.JOB_STATUS_CHANGED, event.event_type)
        self.assertEqual(JobAction.START, edge.scheduling_rule.action)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import shutil
from typing import Text, Dict, Any
from ai_flow.plugin_interface.blob_manager_interface import BlobManager


class LocalBlobManager(BlobManager):
    """
    LocalBlobManager is an implementation of BlobManager based on the local file system.
    LocalBlobManager contains configuration items:
    1. root_directory: The upload directory of the project.
    """
    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        if not self.root_dir:
            raise Exception('`root_directory` option of blob manager config is not configured.')

    def upload(self, local_file_path: Text) -> Text:
        """
        Upload a given file to blob server. Uploaded file will be placed under self.root_dir.

        :param local_file_path: the path of file to be uploaded.
        :return the uri of the uploaded file in blob server.
        """
        if not os.path.exists(self.root_dir):
            os.makedirs(self.root_dir)
        file_name = os.path.basename(local_file_path)
        dest_path = os.path.join(self.root_dir, file_name)
        shutil.move(local_file_path, dest_path)
        return dest_path

    def download(self, remote_file_path: Text, local_dir: Text = None) -> Text:
        """
        Download file from remote blob server to local directory.
        Only files located in self.root_dir can be downloaded by BlobManager.

        :param remote_file_path: The path of file to be downloaded.
        :param local_dir: the local directory.
        :return the local uri of the downloaded file.
        """
        self._check_remote_path_legality(remote_file_path)
        if local_dir is not None:
            file_name = os.path.basename(remote_file_path)
            dest_path = os.path.join(local_dir, file_name)
            if remote_file_path != dest_path:
                shutil.copy(remote_file_path, dest_path)
            return dest_path
        else:
            return remote_file_path

    def _check_remote_path_legality(self, file_path: Text):
        """
        Check if the file can be downloaded by blob manager.

        :param file_path: The path of file to be checked.
        """
        if not file_path.startswith(self.root_dir):
            raise Exception("Cannot download {} from blob server".format(file_path))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import os
import tempfile
import unittest
from unittest.mock import Mock, patch

import cloudpickle

from ai_flow.api.context_extractor import BroadcastAllContextExtractor
from ai_flow.context.project_context import ProjectContext
from ai_flow.project.project_config import ProjectConfig
from ai_flow.workflow.status import Status
from ai_flow.workflow.workflow import Workflow
from ai_flow.workflow.workflow_config import WorkflowConfig
from ai_flow_plugins.scheduler_plugins.airflow.airflow_scheduler import AirFlowScheduler


class TestAirflowScheduler(unittest.TestCase):
    logger = logging.getLogger('TestAirflowScheduler')

    def setUp(self) -> None:
        self.temp_deploy_path = tempfile.TemporaryDirectory().name
        self.logger.info("temp airflow deploy path: {}".format(self.temp_deploy_path))
        self.scheduler = AirFlowScheduler({'airflow_deploy_path': self.temp_deploy_path,
                                           'resource_dir': self.temp_deploy_path,
                                           'notification_server_uri': 'localhost:50051'})
        self.scheduler._airflow_client = Mock()
        self.scheduler.dag_generator = Mock()

    def test_airflow_scheduler_without_dag_deploy_path(self):
        with patch("airflow.settings") as s:
            s.DAGS_FOLDER = 'test_folder'
            scheduler = AirFlowScheduler({'notification_server_uri': 'localhost:50051'})
            self.assertEqual('test_folder', scheduler.config['airflow_deploy_path'])

    def test_airflow_scheduler_submit_workflow(self):
        workflow_name = 'test_workflow'
        workflow = self._get_workflow(workflow_name)

        project_name = 'test_project'
        project_context = self._get_project_context(project_name)

        mock_generated_code = 'mock generated code'
        self.scheduler.dag_generator.generate.return_value = mock_generated_code

        context_extractor = BroadcastAllContextExtractor()
        self.scheduler.submit_workflow(workflow, context_extractor=context_extractor,
                                       project_context=project_context)

        dag_file_path = os.path.join(self.temp_deploy_path, '.'.join([project_name, workflow_name, 'py']))
        self.assertTrue(os.path.exists(dag_file_path))
        with open(dag_file_path, 'rt') as f:
            self.assertEqual(mock_generated_code, f.read())

    def test_airflow_scheduler_delete_workflow(self):
        workflow_name = 'test_workflow'
        workflow = self._get_workflow(workflow_name)

        project_name = 'test_project'
        project_context = self._get_project_context(project_name)

        mock_generated_code = 'mock generated code'
        self.scheduler.dag_generator.generate.return_value = mock_generated_code

        context_extractor = BroadcastAllContextExtractor()
        self.scheduler.submit_workflow(workflow, context_extractor=context_extractor,
                                       project_context=project_context)
        dag_file_path = os.path.join(self.temp_deploy_path, '.'.join([project_name, workflow_name, 'py']))
        self.assertTrue(os.path.exists(dag_file_path))

        with patch('ai_flow_plugins.scheduler_plugins.airflow.airflow_scheduler.'
                   'AirFlowScheduler.pause_workflow_scheduling') as pause_func:
            with patch('ai_flow_plugins.scheduler_plugins.airflow.airflow_scheduler.'
                       'AirFlowScheduler.stop_all_workflow_execution') as stop_func:
                with patch('airflow.api.common.experimental.delete_dag.delete_dag') as delete_func:
                    self.scheduler.delete_workflow(project_name, workflow_name)
        pause_func.assert_called_with(project_name=project_name, workflow_name=workflow_name)
        stop_func.assert_called_with(project_name=project_name, workflow_name=workflow_name)
        delete_func.assert_called_with('{}.{}'.format(project_name, workflow_name))
        self.assertFalse(os.path.exists(dag_file_path))

    def test_airflow_scheduler_submit_workflow_with_customized_context_extractor(self):
        script_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'context_extractor_pickle_maker.py')
        os.system('python {}'.format(script_path))
        context_extractor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                              'context_extractor.pickle')

        try:
            workflow_name = 'test_workflow_with_context_extractor'
            workflow = self._get_workflow(workflow_name)
            with open(context_extractor_path, 'rb') as f:
                context_extractor = cloudpickle.load(f)

            project_name = 'test_project'
            project_context = self._get_project_context(project_name)

            mock_generated_code = 'mock generated code'
            self.scheduler.dag_generator.generate.return_value = mock_generated_code

            self.scheduler.submit_workflow(workflow, context_extractor=context_extractor,
                                           project_context=project_context)

            dag_file_path = os.path.join(self.temp_deploy_path, '.'.join([project_name, workflow_name, 'py']))
            self.assertTrue(os.path.exists(dag_file_path))
            with open(dag_file_path, 'rt') as f:
                self.assertEqual(mock_generated_code, f.read())

        finally:
            os.remove(context_extractor_path)

    @patch('ai_flow_plugins.scheduler_plugins.airflow.airflow_scheduler.create_session')
    def test_airflow_scheduler_kill_workflow_execution(self, mock_session):
        from airflow.models import DagRun
        from airflow.utils.state import State
        mock_session.return_value.__enter__.return_value.query.return_value.filter.return_value.first.return_value = \
            DagRun(dag_id='dag1.run1', state=State.FAILED)

        workflow_execution = self.scheduler.stop_workflow_execution(workflow_execution_id='1|1')
        self.assertEquals(Status.FAILED, workflow_execution.status)

        mock_session.return_value.__enter__.return_value.query.return_value.filter.\
            return_value.first.return_value = DagRun(dag_id='dag1.run1', state=State.RUNNING)
        workflow_execution = self.scheduler.stop_workflow_execution(workflow_execution_id='1|1')
        self.assertEquals(Status.KILLED, workflow_execution.status)

    @staticmethod
    def _get_project_context(project_name):
        project_context: ProjectContext = ProjectContext()
        project_context.project_config = ProjectConfig()
        project_context.project_config.set_project_name(project_name)
        return project_context

    @staticmethod
    def _get_workflow(workflow_name):
        workflow = Workflow()
        workflow.workflow_config = WorkflowConfig(workflow_name=workflow_name)
        return workflow



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import threading
import unittest
from tempfile import TemporaryDirectory
from unittest import mock

from ai_flow.plugin_interface.blob_manager_interface import BlobConfig, BlobManagerFactory

from ai_flow.util.path_util import get_file_dir

from ai_flow_plugins.blob_manager_plugins.s3_blob_manager import S3BlobManager


class TestS3BlobManager(unittest.TestCase):

    @unittest.skipUnless((os.environ.get('blob_server.service_name') is not None
                          and os.environ.get('blob_server.region_name') is not None
                          and os.environ.get('blob_server.api_version') is not None
                          and os.environ.get('blob_server.use_ssl') is not None
                          and os.environ.get('blob_server.verify') is not None
                          and os.environ.get('blob_server.endpoint_url') is not None
                          and os.environ.get('blob_server.access_key_id') is not None
                          and os.environ.get('blob_server.secret_access_key') is not None
                          and os.environ.get('blob_server.bucket_name') is not None), 'need set s3')
    def test_project_upload_download_s3(self):
        upload_file = os.path.abspath(__file__)
        config = {
            'blob_manager_class': 'ai_flow_plugins.blob_manager_plugins.s3_blob_manager.S3BlobManager',
            'blob_manager_config': {
                'service_name': os.environ.get('blob_server.service_name'),
                'region_name': os.environ.get('blob_server.region_name'),
                'api_version': os.environ.get('blob_server.api_version'),
                'use_ssl': os.environ.get('blob_server.use_ssl'),
                'verify': os.environ.get('blob_server.verify'),
                'endpoint_url': os.environ.get('blob_server.endpoint_url'),
                'access_key_id': os.environ.get('blob_server.access_key_id'),
                'secret_access_key': os.environ.get('blob_server.secret_access_key'),
                'bucket_name': os.environ.get('blob_server.bucket_name')
            }
        }

        blob_config = BlobConfig(config)
        blob_manager = BlobManagerFactory.create_blob_manager(
            blob_config.blob_manager_class(), blob_config.blob_manager_config())
        uploaded_path = blob_manager.upload(upload_file)
        file_name = os.path.basename(upload_file)
        self.assertEquals(uploaded_path, file_name)
        with TemporaryDirectory() as tmp_dir:
            downloaded_path = blob_manager.download(file_name, tmp_dir)
            self.assertEqual(os.path.join(tmp_dir, file_name), downloaded_path)

    def test_concurrent_download_s3_file(self):
        project_zip = '/tmp/workflow_1_project.zip'
        if os.path.exists(project_zip):
            os.remove(project_zip)
        config = {'service_name': 's3'}
        s3_blob_manager = S3BlobManager(config)
        zip_file_path = None
        call_count = 0

        def mock_get_s3_object(local_path, object_key):
            nonlocal zip_file_path, call_count
            call_count += 1
            zip_file_path = local_path
            with open(local_path, 'w') as f:
                pass

        s3_blob_manager._get_s3_object = mock_get_s3_object
        with mock.patch(
                'ai_flow_plugins.blob_manager_plugins.blob_manager_utils.extract_project_zip_file'):
            def download_loop():
                for i in range(1000):
                    s3_blob_manager.download('dummy_path', '/tmp')

            try:
                t1 = threading.Thread(target=download_loop)
                t1.start()
                download_loop()
                t1.join()
                self.assertEqual(1, call_count)
            finally:
                if zip_file_path:
                    os.remove(zip_file_path)

    def test__get_s3_object_retry(self):
        config = {'service_name': 's3'}
        s3_blob_manager = S3BlobManager(config)

        with mock.patch.object(s3_blob_manager, 's3_client') as mock_client:
            mock_client.download_fileobj.side_effect = [RuntimeError("boom"), RuntimeError("boom"),
                                                        RuntimeError("boom")]
            with self.assertRaises(RuntimeError):
                s3_blob_manager._get_s3_object('dummy_dest', 'key', retry_sleep_sec=0.1)
                self.assertEqual(3, mock_client.download_fileobj.call_count)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import unittest
import os
import time
import shutil
from ai_flow import AIFlowServerRunner, init_ai_flow_context
from ai_flow.workflow.status import Status
from ai_flow_plugins.job_plugins import flink
from test_flink_processor import SinkWithAddInsertSql, SinkWithExecuteSql, Source, Sink, Transformer, Transformer2, \
    SourceSql, SinkSql
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server
import ai_flow as af

project_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


class TestFlink(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ns_server = start_notification_server()
        config_file = project_path + '/master.yaml'
        cls.master = AIFlowServerRunner(config_file=config_file)
        cls.master.start()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.master.stop()
        generated = '{}/generated'.format(project_path)
        if os.path.exists(generated):
            shutil.rmtree(generated)
        temp = '{}/temp'.format(project_path)
        if os.path.exists(temp):
            shutil.rmtree(temp)
        stop_notification_server(cls.ns_server)

    def setUp(self):
        self.master._clear_db()
        af.current_graph().clear_graph()
        init_ai_flow_context()

    def tearDown(self):
        self.master._clear_db()

    def test_local_flink_task(self):
        flink.set_flink_env(None)
        with af.job_config('task_1'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer())
            af.user_define_operation(input=[processed], processor=Sink())
            flink.set_flink_env(flink.FlinkBatchEnv())
        w = af.workflow_operation.submit_workflow(workflow_name=af.current_workflow_config().workflow_name)
        je = af.workflow_operation.start_job_execution(job_name='task_1', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_1', execution_id='1')
        self.assertEqual(Status.FINISHED, jes[0].status)

    def test_local_flink_task_with_flink_sql_processor(self):
        flink.set_flink_env(None)
        with af.job_config('task_1'):
            af.user_define_operation(processor=SourceSql())
            af.user_define_operation(processor=SinkSql())
            flink.set_flink_env(flink.FlinkStreamEnv())

        workflow = af.workflow_operation.submit_workflow(workflow_name=af.current_workflow_config().workflow_name)
        job_execution = af.workflow_operation.start_job_execution(job_name='task_1', execution_id='1')
        job_executions = af.workflow_operation.get_job_executions(job_name='task_1', execution_id='1')
        self.assertEqual(Status.FINISHED, job_executions[0].status)

    def test_stop_local_flink_task(self):
        with af.job_config('task_1'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer2())
            af.user_define_operation(input=[processed], processor=Sink())
        w = af.workflow_operation.submit_workflow(workflow_name='test_flink')
        je = af.workflow_operation.start_job_execution(job_name='task_1', execution_id='1')
        time.sleep(2)
        af.workflow_operation.stop_job_execution(job_name='task_1', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_1', execution_id='1')
        self.assertEqual(Status.FAILED, jes[0].status)
        self.assertTrue('err' in jes[0].properties)

    def test_local_flink_no_statementset_task(self):
        """
        This is a test case for issue 475.
        https://github.com/alibaba/flink-ai-extended/issues/475
        It calls 'TableEnv.execute_sql' in FlinkPythonProcessor and no need to call 
        StatementSet.execute().
        flink_run_main.py should be able to handle this case.
        """
        with af.job_config('task_1'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer())
            af.user_define_operation(input=[processed], processor=SinkWithExecuteSql())
        af.workflow_operation.submit_workflow(
            workflow_name=af.current_workflow_config().workflow_name)
        af.workflow_operation.start_job_execution(job_name='task_1', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_1', execution_id='1')
        self.assertEqual(Status.FINISHED, jes[0].status)

    def test_local_flink_execute_statementset_add_insert_sql_task(self):
        """
        This is a test case for issue 475.
        https://github.com/alibaba/flink-ai-extended/issues/475
        It calls 'Statementset.add_insert_sql' in FlinkPythonProcessor and need to call 
        StatementSet.execute().
        flink_run_main.py should be able to handle this case.
        """
        with af.job_config('task_1'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(
                input=[input_example], transform_processor=Transformer())
            af.user_define_operation(
                input=[processed], processor=SinkWithAddInsertSql())
        af.workflow_operation.submit_workflow(
            workflow_name=af.current_workflow_config().workflow_name)
        af.workflow_operation.start_job_execution(
            job_name='task_1', execution_id='1')
        jes = af.workflow_operation.get_job_executions(
            job_name='task_1', execution_id='1')
        self.assertEqual(Status.FINISHED, jes[0].status)

    @unittest.skip("need start flink cluster")
    def test_cluster_flink_task(self):
        with af.job_config('task_2'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer())
            af.user_define_operation(input=[processed], processor=Sink())
        w = af.workflow_operation.submit_workflow(workflow_name=af.current_workflow_config().workflow_name)
        je = af.workflow_operation.start_job_execution(job_name='task_2', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_2', execution_id='1')
        self.assertEqual(Status.FINISHED, jes[0].status)

    @unittest.skip("need start flink cluster")
    def test_cluster_stop_local_flink_task(self):
        with af.job_config('task_2'):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer2())
            af.user_define_operation(input=[processed], processor=Sink())
        w = af.workflow_operation.submit_workflow(workflow_name='test_python')
        je = af.workflow_operation.start_job_execution(job_name='task_2', execution_id='1')
        time.sleep(20)
        af.workflow_operation.stop_job_execution(job_name='task_2', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_2', execution_id='1')
        self.assertEqual(Status.FAILED, jes[0].status)
        self.assertTrue('err' in jes[0].properties)

    @unittest.skip("need start flink cluster")
    def test_cluster_stop_local_flink_task_with_cancel(self):
        job_name = 'task_3'
        with af.job_config(job_name):
            input_example = af.user_define_operation(processor=Source())
            processed = af.transform(input=[input_example], transform_processor=Transformer2())
            af.user_define_operation(input=[processed], processor=Sink())
        w = af.workflow_operation.submit_workflow(workflow_name='test_python')
        je = af.workflow_operation.start_job_execution(job_name=job_name, execution_id='1')
        time.sleep(10)
        af.workflow_operation.stop_job_execution(job_name=job_name, execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name=job_name, execution_id='1')
        self.assertEqual(Status.FAILED, jes[0].status)
        self.assertTrue('err' in jes[0].properties)

    @unittest.skip("need start flink cluster and run (nc -lk 9999)")
    def test_cluster_stop_local_flink_task_with_stop(self):
        job_name = 'task_4'
        flink_home = os.environ.get('FLINK_HOME')
        word_count_jar = os.path.join(flink_home, 'examples', 'streaming', 'SocketWindowWordCount.jar')
        jar_dir = os.path.join(project_path, 'dependencies', 'jar')
        if not os.path.exists(jar_dir):
            os.makedirs(jar_dir)
            shutil.copy(word_count_jar, jar_dir)

        args = ['--port', '9999']
        with af.job_config(job_name):
            af.user_define_operation(processor=flink.FlinkJavaProcessor(entry_class=None,
                                                                        main_jar_file='SocketWindowWordCount.jar',
                                                                        args=args))
        w = af.workflow_operation.submit_workflow(workflow_name=af.current_workflow_config().workflow_name)
        je = af.workflow_operation.start_job_execution(job_name=job_name, execution_id='1')
        time.sleep(10)
        af.workflow_operation.stop_job_execution(job_name=job_name, execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name=job_name, execution_id='1')
        self.assertEqual(Status.FAILED, jes[0].status)
        self.assertTrue('err' in jes[0].properties)
        dep_dir = os.path.join(project_path, 'dependencies')
        if os.path.exists(dep_dir):
            shutil.rmtree(dep_dir)
        savepoint_path = '/tmp/savepoint_test_cluster_stop_local_flink_task_with_stop'
        self.assertTrue(os.path.exists(savepoint_path))
        shutil.rmtree(savepoint_path)

    @unittest.skip("need start flink cluster")
    def test_cluster_flink_java_task(self):
        flink_home = os.environ.get('FLINK_HOME')
        word_count_jar = os.path.join(flink_home, 'examples', 'batch', 'WordCount.jar')
        output_file = os.path.join(flink_home, 'log', 'output')
        if os.path.exists(output_file):
            os.remove(output_file)
        jar_dir = os.path.join(project_path, 'dependencies', 'jar')
        if not os.path.exists(jar_dir):
            os.makedirs(jar_dir)
            shutil.copy(word_count_jar, jar_dir)

        args = ['--input', os.path.join(flink_home, 'conf', 'flink-conf.yaml'), '--output', output_file]
        with af.job_config('task_2'):
            af.user_define_operation(processor=flink.FlinkJavaProcessor(entry_class=None,
                                                                        main_jar_file='WordCount.jar',
                                                                        args=args))
        w = af.workflow_operation.submit_workflow(workflow_name=af.current_workflow_config().workflow_name)
        je = af.workflow_operation.start_job_execution(job_name='task_2', execution_id='1')
        jes = af.workflow_operation.get_job_executions(job_name='task_2', execution_id='1')
        self.assertEqual(Status.FINISHED, jes[0].status)
        dep_dir = os.path.join(project_path, 'dependencies')
        if os.path.exists(dep_dir):
            shutil.rmtree(dep_dir)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Text
from kubernetes import config

kubernetes_cluster_available = False


def load_kubernetes_config(config_file_path: Text = None):
    global kubernetes_cluster_available
    if config_file_path is not None:
        # load kubernetes config from user provided config file
        config.load_kube_config(config_file=config_file_path)
        kubernetes_cluster_available = True
    else:
        try:
            # load local kubernetes config
            config.load_kube_config()
            kubernetes_cluster_available = True
        except Exception as e:
            try:
                config.load_incluster_config()
                kubernetes_cluster_available = True
            except Exception as e:
                pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest

from ai_flow_plugins.job_plugins.bash.bash_processor import BashProcessor

from ai_flow.ai_graph.ai_node import AINode

from ai_flow.workflow.job_config import JobConfig

from ai_flow.ai_graph.ai_graph import AISubGraph
from ai_flow_plugins.job_plugins.python import PythonProcessor

from ai_flow_plugins.job_plugins.python.python_job_plugin import PythonJobGenerator, PythonJob


class TestPythonJobGenerator(unittest.TestCase):
    def test_generate_invalid_subgraph_raise_exception(self):
        sub_graph = AISubGraph(JobConfig())
        ai_node = AINode(processor=BashProcessor('hello'))
        sub_graph.add_node(ai_node)
        sub_graph.add_node(AINode(processor=PythonProcessor()))
        python_job_generator = PythonJobGenerator()

        with self.assertRaises(Exception):
            python_job_generator.generate(sub_graph)

    def test_generate(self):
        sub_graph = AISubGraph(JobConfig())
        sub_graph.add_node(AINode(processor=PythonProcessor()))
        sub_graph.add_node(AINode(processor=PythonProcessor()))
        python_job_generator = PythonJobGenerator()

        job = python_job_generator.generate(sub_graph)
        self.assertIsInstance(job, PythonJob)




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.model_center.entity._model_repo_entity import _ModelRepoEntity


class ModelVersion(_ModelRepoEntity):
    """
    AIFlow entity for Model Version. A model version is uniquely identified using underlying
    :py:class:`ai_flow.model_center.entity.RegisteredModel` and user-defined version.
    """

    def __init__(self, model_name, model_version):
        """
        Construct a :py:class:`ai_flow.model_center.entity.ModelVersion` instance

        :param model_name: Unique name for this registered model within Model Registry.
        :param model_version: Int, the model version.
        """
        super(ModelVersion, self).__init__()
        self.model_name = model_name
        self.model_version = model_version

    # proto mappers
    @classmethod
    def from_proto(cls, proto):
        model_meta = proto.model_meta
        return ModelVersion(model_meta.model_name.value if model_meta.HasField("model_name") else None,
                            model_meta.model_version.value if model_meta.HasField("model_version") else None)

    @classmethod
    def from_resp_proto(cls, proto):
        return ModelVersion(proto.model_name.value if proto.HasField("model_name") else None,
                            proto.model_version.value if proto.HasField("model_version") else None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Any, Dict

import requests

from airflow.decorators import dag, task
from airflow.models.baseoperator import BaseOperator
from airflow.operators.email import EmailOperator
from airflow.utils.dates import days_ago

DEFAULT_ARGS = {"owner": "airflow"}


class GetRequestOperator(BaseOperator):
    """Custom operator to sand GET request to provided url"""

    def __init__(self, *, url: str, **kwargs):
        super().__init__(**kwargs)
        self.url = url

    def execute(self, context):
        return requests.get(self.url).json()


# [START dag_decorator_usage]
@dag(default_args=DEFAULT_ARGS, schedule_interval=None, start_date=days_ago(2))
def example_dag_decorator(email: str = 'example@example.com'):
    """
    DAG to send server IP to email.

    :param email: Email to send IP to. Defaults to example@example.com.
    :type email: str
    """
    get_ip = GetRequestOperator(task_id='get_ip', url="http://httpbin.org/get")

    @task(multiple_outputs=True)
    def prepare_email(raw_json: Dict[str, Any]) -> Dict[str, str]:
        external_ip = raw_json['origin']
        return {
            'subject': f'Server connected from {external_ip}',
            'body': f'Seems like today your server executing Airflow is connected from IP {external_ip}<br>',
        }

    email_info = prepare_email(get_ip.output)

    EmailOperator(
        task_id='send_email', to=email, subject=email_info['subject'], html_content=email_info['body']
    )


dag = example_dag_decorator()
# [END dag_decorator_usage]
